# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from six.moves import range

import tempfile

from django.apps import apps
from django.db.migrations.executor import MigrationExecutor
from django.db import connection
from aiida.backends.testbase import AiidaTestCase
from aiida.common.exceptions import IntegrityError
from aiida.manage.database.integrity import deduplicate_node_uuids, verify_node_uuid_uniqueness


class TestMigrations(AiidaTestCase):

    @property
    def app(self):
        return apps.get_containing_app_config(type(self).__module__).name.split('.')[-1]

    migrate_from = None
    migrate_to = None

    def setUp(self):
        assert self.migrate_from and self.migrate_to, \
            "TestCase '{}' must define migrate_from and migrate_to properties".format(type(self).__name__)
        self.migrate_from = [(self.app, self.migrate_from)]
        self.migrate_to = [(self.app, self.migrate_to)]
        executor = MigrationExecutor(connection)
        old_apps = executor.loader.project_state(self.migrate_from).apps

        # Reverse to the original migration
        executor.migrate(self.migrate_from)

        self.setUpBeforeMigration(old_apps)

        # Run the migration to test
        executor = MigrationExecutor(connection)
        executor.loader.build_graph()
        executor.migrate(self.migrate_to)

        self.apps = executor.loader.project_state(self.migrate_to).apps

    def setUpBeforeMigration(self, apps):
        pass


class TestDuplicateNodeUuidMigration(TestMigrations):

    migrate_from = '0013_django_1_8'
    migrate_to = '0014_add_node_uuid_unique_constraint'

    def setUpBeforeMigration(self, apps):
        from aiida.orm.data.bool import Bool
        from aiida.orm.data.int import Int

        self.file_name = 'test.temp'
        self.file_content = '#!/bin/bash\n\necho test run\n'

        with tempfile.NamedTemporaryFile(mode='w+') as handle:
            handle.write(self.file_content)
            handle.flush()

            self.nodes_boolean = []
            self.nodes_integer = []
            self.n_bool_duplicates = 2
            self.n_int_duplicates = 4

            node_bool = Bool(True)
            node_bool.add_path(handle.name, self.file_name)
            node_bool.store()

            node_int = Int(1)
            node_int.add_path(handle.name, self.file_name)
            node_int.store()

            self.nodes_boolean.append(node_bool)
            self.nodes_integer.append(node_int)

            for i in range(self.n_bool_duplicates):
                node = Bool(True)
                node.dbnode.uuid = node_bool.uuid
                node.add_path(handle.name, self.file_name)
                node.store()
                self.nodes_boolean.append(node)

            for i in range(self.n_int_duplicates):
                node = Int(1)
                node.dbnode.uuid = node_int.uuid
                node.add_path(handle.name, self.file_name)
                node.store()
                self.nodes_integer.append(node)

        # Verify that there are duplicate UUIDs by checking that the following function raises
        with self.assertRaises(IntegrityError):
            verify_node_uuid_uniqueness()

        # Now run the function responsible for solving duplicate UUIDs which would also be called by the user
        # through the `verdi database integrity duplicate-node-uuid` command
        deduplicate_node_uuids(dry_run=False)

    def test_deduplicated_uuids(self):
        """Verify that after the migration, all expected nodes are still there with unique UUIDs."""
        from aiida.orm import load_node

        # If the duplicate UUIDs were successfully fixed, the following should not raise.
        verify_node_uuid_uniqueness()

        # Reload the nodes by PK and check that all UUIDs are now unique
        nodes_boolean = [load_node(node.pk) for node in self.nodes_boolean]
        uuids_boolean = [node.uuid for node in nodes_boolean]
        self.assertEqual(len(set(uuids_boolean)), len(nodes_boolean))

        nodes_integer = [load_node(node.pk) for node in self.nodes_integer]
        uuids_integer = [node.uuid for node in nodes_integer]
        self.assertEqual(len(set(uuids_integer)), len(nodes_integer))

        for node in nodes_boolean:
            with node._get_folder_pathsubfolder.open(self.file_name) as handle:
                content = handle.read()
                self.assertEqual(content, self.file_content)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=superfluous-parens,too-many-locals,too-many-statements,too-many-branches
"""Functions to create, inspect and manipulate profiles in the configuration file."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
import sys

from aiida.cmdline.utils import echo


def setup_profile(profile, only_config, set_default=False, non_interactive=False, **kwargs):
    """
    Setup an AiiDA profile and AiiDA user (and the AiiDA default user).

    :param profile: Profile name
    :param only_config: do not create a new user
    :param set_default: set the new profile as the default
    :param non_interactive: do not prompt for configuration values, fail if not all values are given as kwargs.
    :param backend: one of 'django', 'sqlalchemy'
    :param email: valid email address for the user
    :param db_host: hostname for the database
    :param db_port: port to connect to the database
    :param db_user: name of the db user
    :param db_pass: password of the db user
    """
    from aiida.backends import settings
    from aiida.backends.profile import BACKEND_SQLA, BACKEND_DJANGO
    from aiida.backends.utils import set_backend_type
    from aiida.cmdline import EXECNAME
    from aiida.cmdline.commands import cmd_user
    from aiida.common.exceptions import InvalidOperation
    from aiida.common.setup import (create_base_dirs, create_configuration, set_default_profile, DEFAULT_UMASK,
                                    create_config_noninteractive)

    only_user_config = only_config

    # Create the directories to store the configuration files
    create_base_dirs()
    if settings.AIIDADB_PROFILE and profile:
        sys.exit('the profile argument cannot be used if verdi is called with -p option: {} and {}'.format(
            settings.AIIDADB_PROFILE, profile))
    gprofile = settings.AIIDADB_PROFILE or profile
    if gprofile == profile:
        settings.AIIDADB_PROFILE = profile
    if not settings.AIIDADB_PROFILE:
        settings.AIIDADB_PROFILE = 'default'

    # used internally later
    gprofile = settings.AIIDADB_PROFILE

    created_conf = None
    # ask and store the configuration of the DB
    if non_interactive:
        try:
            created_conf = create_config_noninteractive(
                profile=gprofile,
                backend=kwargs['backend'],
                email=kwargs['email'],
                db_host=kwargs['db_host'],
                db_port=kwargs['db_port'],
                db_name=kwargs['db_name'],
                db_user=kwargs['db_user'],
                db_pass=kwargs.get('db_pass', ''),
                repo=kwargs['repo'],
                force_overwrite=kwargs.get('force_overwrite', False))
        except ValueError as exception:
            echo.echo_critical("Error during configuation: {}".format(exception))
        except KeyError as exception:
            import traceback
            echo.echo(traceback.format_exc())
            echo.echo_critical(
                "--non-interactive requires all values to be given on the commandline! Missing argument: {}".format(
                    exception.args[0]))
    else:
        try:
            created_conf = create_configuration(profile=gprofile)
        except ValueError as exception:
            echo.echo_critical("Error during configuration: {}".format(exception))

        # Set default DB profile
        set_default_profile(gprofile, force_rewrite=False)

    if only_user_config:
        echo.echo("Only user configuration requested, skipping the migrate command")
    else:
        echo.echo("Executing now a migrate command...")

        backend_choice = created_conf['AIIDADB_BACKEND']
        if backend_choice == BACKEND_DJANGO:
            echo.echo("...for Django backend")
            # The correct profile is selected within load_dbenv.
            # Setting os.umask here since sqlite database gets created in
            # this step.
            old_umask = os.umask(DEFAULT_UMASK)

            # This check should be done more properly
            # try:
            #     backend_type = get_backend_type()
            # except KeyError:
            #     backend_type = None
            #
            # if backend_type is not None and backend_type != BACKEND_DJANGO:
            #     raise InvalidOperation("An already existing database found"
            #                            "and a different than the selected"
            #                            "backend was used for its "
            #                            "management.")

            try:
                from aiida.backends.djsite.utils import pass_to_django_manage
                pass_to_django_manage([EXECNAME, 'migrate'], profile=gprofile)
            finally:
                os.umask(old_umask)

            set_backend_type(BACKEND_DJANGO)

        elif backend_choice == BACKEND_SQLA:
            echo.echo("...for SQLAlchemy backend")
            from aiida import is_dbenv_loaded
            from aiida.backends.sqlalchemy.utils import _load_dbenv_noschemacheck, check_schema_version
            from aiida.backends.profile import load_profile

            # We avoid calling load_dbenv since we want to force the schema
            # migration
            if not is_dbenv_loaded():
                settings.LOAD_DBENV_CALLED = True
                # This is going to set global variables in settings, including settings.BACKEND
                load_profile()
                _load_dbenv_noschemacheck()

            # Perform the needed migration quietly
            check_schema_version(force_migration=True)
            set_backend_type(BACKEND_SQLA)

        else:
            raise InvalidOperation("Not supported backend selected.")

    echo.echo("Database was created successfully")

    # I create here the default user
    echo.echo("Loading new environment...")
    if only_user_config:
        from aiida.backends.utils import load_dbenv, is_dbenv_loaded
        # db environment has not been loaded in this case
        if not is_dbenv_loaded():
            load_dbenv()

    from aiida.common.setup import DEFAULT_AIIDA_USER
    from aiida.orm.backend import construct_backend

    backend = construct_backend()
    if not backend.users.find(email=DEFAULT_AIIDA_USER):
        echo.echo("Installing default AiiDA user...")
        nuser = backend.users.create(email=DEFAULT_AIIDA_USER, first_name="AiiDA", last_name="Daemon")
        nuser.is_active = True
        nuser.store()

    from aiida.common.utils import get_configured_user_email
    email = get_configured_user_email()
    echo.echo("Starting user configuration for {}...".format(email))
    if email == DEFAULT_AIIDA_USER:
        echo.echo("You set up AiiDA using the default Daemon email ({}),".format(email))
        echo.echo("therefore no further user configuration will be asked.")
    else:
        if non_interactive:
            # Here we map the keyword arguments onto the command line arguments
            # for verdi user configure.  We have to be careful that there the
            # argument names are the same as those int he kwargs dict
            commands = [kwargs['email'], '--non-interactive']

            for arg in ('first_name', 'last_name', 'institution'):
                value = kwargs.get(arg, None)
                if value is not None:
                    commands.extend(('--{}'.format(arg.replace('_', '-')), str(value)))
        else:
            commands = [email]

        # Ask to configure the user
        try:
            # pylint: disable=no-value-for-parameter
            cmd_user.configure(commands)
        except SystemExit:
            # Have to catch this as the configure command will do a sys.exit()
            pass

    if set_default:
        set_default_profile(profile, force_rewrite=True)

    echo.echo("Setup finished.")



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os

from aiida.backends import settings
from aiida.common import setup


CONFIG_DIR = setup.AIIDA_CONFIG_FOLDER
DAEMON_DIR = 'daemon'
DAEMON_LOG_DIR = 'daemon/log'

CIRCUS_PID_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.pid')
DAEMON_PID_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'aiida-{}.pid')
CIRCUS_LOG_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_LOG_DIR, 'circus-{}.log')
DAEMON_LOG_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_LOG_DIR, 'aiida-{}.log')
CIRCUS_PORT_FILE_TEMPLATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.port')
CIRCUS_SOCKET_FILE_TEMPATE = os.path.join(CONFIG_DIR, DAEMON_DIR, 'circus-{}.sockets')
CIRCUS_CONTROLLER_SOCKET_TEMPLATE = 'circus.c.sock'
CIRCUS_PUBSUB_SOCKET_TEMPLATE = 'circus.p.sock'
CIRCUS_STATS_SOCKET_TEMPLATE = 'circus.s.sock'


def get_current_profile_name():
    """
    Return the currently configured profile name or if not set, the default profile
    """
    return settings.AIIDADB_PROFILE or setup.get_default_profile_name()


def get_profile_config(name=None):
    """
    Return the configuration of the currently active profile or if not set, the default profile

    :param name: the profile name whose configuration to return
    :return: the dictionary containing the profile configuration
    :raises MissingConfigurationError: if the configuration file cannot be found
    :raises ProfileConfigurationError: if the name is not found in the configuration file
    """
    from aiida.common.exceptions import MissingConfigurationError, ProfileConfigurationError
    from aiida.common.setup import get_config

    if name is None:
        name = get_current_profile_name()

    try:
        config = get_config()
    except MissingConfigurationError:
        raise MissingConfigurationError('could not load the configuration file')

    try:
        profile = config['profiles'][name]
    except KeyError:
        raise ProfileConfigurationError('invalid profile name "{}"'.format(name))

    return profile


def get_profile(name=None):
    """
    Return the profile for the given name or the default one if not specified.

    :param name: the profile name, will use the default profile if None
    :return: the profile
    :rtype: :class:`aiida.common.profile.Profile`
    :raises MissingConfigurationError: if the configuration file cannot be found
    :raises ProfileConfigurationError: if the name is not found in the configuration file
    """
    if name is None:
        name = get_current_profile_name()

    config = get_profile_config(name)
    return Profile(name, config)


class Profile(object):
    """
    Convenience class that loads the current profile name and corresponding configuration and
    can subsequently provide profile specific information
    """

    _RMQ_PREFIX = 'aiida-{uuid}'

    def __init__(self, name, config):
        self._name = name
        self._config = config

        # Currently, whether a profile is a test profile is solely determined by its name starting with 'test_'
        if self.name.startswith('test_'):
            self._test_profile = True
        else:
            self._test_profile = False

    @property
    def config(self):
        return self._config

    @property
    def name(self):
        return self._name

    @property
    def uuid(self):
        return self.config[setup.PROFILE_UUID_KEY]

    @property
    def rmq_prefix(self):
        return self._RMQ_PREFIX.format(uuid=self.uuid)

    @property
    def is_test_profile(self):
        return self._test_profile

    @property
    def filepaths(self):
        return {
            'circus': {
                'log': CIRCUS_LOG_FILE_TEMPLATE.format(self.name),
                'pid': CIRCUS_PID_FILE_TEMPLATE.format(self.name),
                'port': CIRCUS_PORT_FILE_TEMPLATE.format(self.name),
                'socket': {
                    'file': CIRCUS_SOCKET_FILE_TEMPATE.format(self.name),
                    'controller': CIRCUS_CONTROLLER_SOCKET_TEMPLATE,
                    'pubsub': CIRCUS_PUBSUB_SOCKET_TEMPLATE,
                    'stats': CIRCUS_STATS_SOCKET_TEMPLATE,
                }
            },
            'daemon': {
                'log': DAEMON_LOG_FILE_TEMPLATE.format(self.name),
                'pid': DAEMON_PID_FILE_TEMPLATE.format(self.name),
            }
        }



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import abc
from abc import abstractproperty, ABCMeta
import six

__all__ = ['Backend', 'construct_backend', 'Collection', 'CollectionEntry']

_DJANGO_BACKEND = None
_SQLA_BACKEND = None


def construct_backend(backend_type=None):
    """
    Construct a concrete backend instance based on the backend_type or use the global backend value if not specified.

    :param backend_type: get a backend instance based on the specified type (or default)
    :return: :class:`aiida.orm.backend.Backend`
    """
    if backend_type is None:
        from aiida.backends import settings
        backend_type = settings.BACKEND

    if backend_type == 'django':
        global _DJANGO_BACKEND
        if _DJANGO_BACKEND is None:
            from aiida.orm.implementation.django.backend import DjangoBackend
            _DJANGO_BACKEND = DjangoBackend()
        return _DJANGO_BACKEND
    elif backend_type == 'sqlalchemy':
        global _SQLA_BACKEND
        if _SQLA_BACKEND is None:
            from aiida.orm.implementation.sqlalchemy.backend import SqlaBackend
            _SQLA_BACKEND = SqlaBackend()
        return _SQLA_BACKEND
    else:
        raise ValueError("The specified backend {} is currently not implemented".format(backend_type))


@six.add_metaclass(abc.ABCMeta)
class Backend(object):
    """The public interface that defines a backend factory that creates backend specific concrete objects."""

    @abc.abstractproperty
    def logs(self):
        """
        Return the collection of log entries

        :return: the log collection
        :rtype: :class:`aiida.orm.log.Log`
        """

    @abc.abstractproperty
    def users(self):
        """
        Return the collection of users

        :return: the users collection
        :rtype: :class:`aiida.orm.user.UserCollection`
        """

    @abc.abstractproperty
    def authinfos(self):
        """
        Return the collection of authorisation information objects

        :return: the authinfo collection
        :rtype: :class:`aiida.orm.authinfo.AuthInfoCollection`
        """

    @abc.abstractproperty
    def computers(self):
        """
        Return the collection of computer objects

        :return: the computers collection
        :rtype: :class:`aiida.orm.computer.ComputerCollection`
        """

    @abc.abstractproperty
    def query_manager(self):
        """
        Return the query manager for the objects stored in the backend

        :return: the query manger
        :rtype: :class:`aiida.backends.general.abstractqueries.AbstractQueryManager`
        """

    @abc.abstractmethod
    def query_builder(self):
        """
        Construct a new instance of a QueryBuilder for this backend
        :return: the new query builder
        """


class Collection(object):
    """Container class that represents a collection of entries of a particular backend entity."""

    def __init__(self, backend):
        self._backend = backend

    @property
    def backend(self):
        """Return the backend."""
        return self._backend


class CollectionEntry(object):
    """Class that represents an entry within a collection of entries of a particular backend entity."""

    def __init__(self, backend):
        """
        :param backend: The backend instance
        :type backend: :class:`aiida.orm.backend.Backend`
        """
        if not isinstance(backend, Backend):
            raise TypeError("Must supply a backend, got '{}'".format(type(backend)))
        self._backend = backend

    @property
    def backend(self):
        """Return the backend."""
        return self._backend



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.backends.testbase import AiidaTestCase
from aiida.orm import CalculationFactory
from aiida.orm.data.parameter import ParameterData
from aiida.orm.data.bool import Bool
from aiida.orm.data.float import Float
from aiida.orm.data.int import Int
from aiida.work import WorkChain, Process

DEFAULT_INT = 256


class TestWorkChain(WorkChain):
    @classmethod
    def define(cls, spec):
        super(TestWorkChain, cls).define(spec)
        spec.input('a', valid_type=Int, help='A pointless integer')
        spec.input('b', valid_type=Float, help='A pointless float')
        spec.input('c.d', valid_type=Bool, help='A pointless boolean')
        spec.input('e', valid_type=Int, default=Int(DEFAULT_INT))


class TestProcessBuilder(AiidaTestCase):

    def setUp(self):
        super(TestProcessBuilder, self).setUp()
        self.assertIsNone(Process.current())
        self.calculation_class = CalculationFactory('simpleplugins.templatereplacer')
        self.process_class = self.calculation_class.process()
        self.builder = self.process_class.get_builder()

    def tearDown(self):
        super(TestProcessBuilder, self).tearDown()
        self.assertIsNone(Process.current())

    def test_process_builder_attributes(self):
        """
        Check that the builder has all the input ports of the process class as attributes
        """
        for name, port in self.process_class.spec().inputs.items():
            self.assertTrue(hasattr(self.builder, name))

    def test_process_builder_set_attributes(self):
        """
        Verify that setting attributes in builder works
        """
        label = 'Test label'
        description = 'Test description'

        self.builder.label = label
        self.builder.description = description

        self.assertEquals(self.builder.label, label)
        self.assertEquals(self.builder.description, description)

    def test_workchain(self):
        """
        Verify that the attributes of the TestWorkChain can be set but defaults are not there
        """
        builder = TestWorkChain.get_builder()
        builder.a = Int(2)
        builder.b = Float(2.3)
        builder.c.d = Bool(True)
        self.assertEquals(builder, {'a': Int(2), 'b': Float(2.3), 'c': {'d': Bool(True)}})

    def test_invalid_setattr_raises(self):
        """
        Verify that __setattr__ cannot be called on a terminal Port
        """
        builder = TestWorkChain.get_builder()
        with self.assertRaises(AttributeError):
            builder.a.b = 3

    def test_dynamic_getters_value(self):
        """
        Verify that getters will return the actual value
        """
        builder = TestWorkChain.get_builder()
        builder.a = Int(2)
        builder.b = Float(2.3)
        builder.c.d = Bool(True)

        # Verify that the correct type is returned by the getter
        self.assertTrue(isinstance(builder.a, Int))
        self.assertTrue(isinstance(builder.b, Float))
        self.assertTrue(isinstance(builder.c.d, Bool))

        # Verify that the correct value is returned by the getter
        self.assertEquals(builder.a, Int(2))
        self.assertEquals(builder.b, Float(2.3))
        self.assertEquals(builder.c.d, Bool(True))

    def test_dynamic_getters_doc_string(self):
        """
        Verify that getters have the correct docstring
        """
        builder = TestWorkChain.get_builder()
        self.assertEquals(builder.__class__.a.__doc__, str(TestWorkChain.spec().inputs['a']))
        self.assertEquals(builder.__class__.b.__doc__, str(TestWorkChain.spec().inputs['b']))
        self.assertEquals(builder.__class__.c.__doc__, str(TestWorkChain.spec().inputs['c']))
        self.assertEquals(builder.c.__class__.d.__doc__, str(TestWorkChain.spec().inputs['c']['d']))

    def test_job_calculation_get_builder_restart(self):
        """
        Test the get_builder_restart method of JobCalculation class
        """
        from aiida.orm.calculation.job import JobCalculation

        original = JobCalculation()
        original.set_option('resources', {'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        original.set_option('max_wallclock_seconds', 1800)
        original.set_computer(self.computer)
        original.label = 'original'
        original.store()

        builder = original.get_builder_restart()

        self.assertDictEqual(builder.options, original.get_options(only_actually_set=True))

    def test_code_get_builder(self):
        """
        Test that the get_builder method of Code returns a builder
        where the code is already set.
        """
        from aiida.orm import Code

        code1 = Code()
        # This also sets the code as a remote code
        code1.set_remote_computer_exec((self.computer, '/bin/true'))
        code1.label = 'test_code1'
        code1.set_input_plugin_name('simpleplugins.templatereplacer')
        code1.store()

        # Check that I can get a builder
        builder = code1.get_builder()
        self.assertEquals(builder.code.pk, code1.pk)

        # Check that I can set the parameters
        builder.parameters = ParameterData(dict={})

        # Check that it complains for an unknown input
        with self.assertRaises(AttributeError):
            builder.unknown_parameter = 3

        # Check that it complains if the type is not the correct one
        # (for the simpleplugins.templatereplacer, it should be a
        # ParameterData)
        with self.assertRaises(ValueError):
            builder.parameters = Int(3)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Test suite to test the `verdi profile` commands."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from click.testing import CliRunner
from six.moves import range

from aiida.backends.testbase import AiidaTestCase
from aiida.common import setup as aiida_cfg


class TestVerdiProfileSetup(AiidaTestCase):
    """
    Test suite to test verdi profile command
    """

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create dummy 5 profiles and set first profile as default profile.
        All tests will run on these dummy profiles.

        :param args: list of arguments
        :param kwargs: list of keyword arguments
        """
        super(TestVerdiProfileSetup, cls).setUpClass()

        import tempfile

        cls._old_aiida_config_folder = None
        cls._new_aiida_config_folder = tempfile.mkdtemp()

        cls._old_aiida_config_folder = aiida_cfg.AIIDA_CONFIG_FOLDER
        aiida_cfg.AIIDA_CONFIG_FOLDER = cls._new_aiida_config_folder
        aiida_cfg.create_base_dirs()

        dummy_profile_list = ['dummy_profile1', 'dummy_profile2', 'dummy_profile3', 'dummy_profile4', 'dummy_profile5']
        dummy_profile_list = ["{}_{}".format(profile_name, get_random_string(4)) for profile_name in dummy_profile_list]

        for profile_name in dummy_profile_list:
            dummy_profile = {}
            dummy_profile['backend'] = 'django'
            dummy_profile['db_host'] = 'localhost'
            dummy_profile['db_port'] = '5432'
            dummy_profile['email'] = 'dummy@localhost'
            dummy_profile['db_name'] = profile_name
            dummy_profile['db_user'] = 'dummy_user'
            dummy_profile['db_pass'] = 'dummy_pass'
            dummy_profile['repo'] = aiida_cfg.AIIDA_CONFIG_FOLDER + '/repository_' + profile_name
            aiida_cfg.create_config_noninteractive(profile=profile_name, **dummy_profile)

        aiida_cfg.set_default_profile(dummy_profile_list[0], force_rewrite=True)
        cls.dummy_profile_list = dummy_profile_list

    @classmethod
    def tearDownClass(cls, *args, **kwargs):
        """
        Remove dummy profiles created in setUpClass.

        :param args: list of arguments
        :param kwargs: list of keyword arguments
        """
        import os
        import shutil

        aiida_cfg.AIIDA_CONFIG_FOLDER = cls._old_aiida_config_folder

        if os.path.isdir(cls._new_aiida_config_folder):
            shutil.rmtree(cls._new_aiida_config_folder)

    def setUp(self):
        """
        Create runner object to run tests
        """
        self.runner = CliRunner()

    def test_help(self):
        """
        Tests help text for all profile sub commands
        """
        options = ["--help"]
        from aiida.cmdline.commands.cmd_profile import (profile_list, profile_setdefault, profile_delete)

        result = self.runner.invoke(profile_list, options)
        self.assertIsNone(result.exception)
        self.assertIn('Usage', result.output)

        result = self.runner.invoke(profile_setdefault, options)
        self.assertIsNone(result.exception)
        self.assertIn('Usage', result.output)

        result = self.runner.invoke(profile_delete, options)
        self.assertIsNone(result.exception)
        self.assertIn('Usage', result.output)

    def test_list(self):
        """
        Test for verdi profile list command
        """
        from aiida.cmdline.commands.cmd_profile import profile_list
        result = self.runner.invoke(profile_list)
        self.assertIsNone(result.exception)
        self.assertIn('configuration folder: ' + self._new_aiida_config_folder, result.output)
        self.assertIn('* {}'.format(self.dummy_profile_list[0]), result.output)
        self.assertIn(self.dummy_profile_list[1], result.output)

    def test_setdefault(self):
        """
        Test for verdi profile setdefault command
        """
        from aiida.cmdline.commands.cmd_profile import profile_setdefault
        result = self.runner.invoke(profile_setdefault, [self.dummy_profile_list[1]])
        self.assertIsNone(result.exception)

        from aiida.cmdline.commands.cmd_profile import profile_list
        result = self.runner.invoke(profile_list)

        self.assertIsNone(result.exception)
        self.assertIn('configuration folder: ' + self._new_aiida_config_folder, result.output)
        self.assertIn('* {}'.format(self.dummy_profile_list[1]), result.output)
        self.assertIsNone(result.exception)

    def test_show(self):
        """
        Test for verdi profile show command
        """
        from aiida.cmdline.commands.cmd_profile import profile_show
        from aiida.common.setup import get_config

        config = get_config()
        profiles = config['profiles']
        profile_name = self.dummy_profile_list[0]
        profile = profiles[profile_name]

        result = self.runner.invoke(profile_show, [profile_name])
        self.assertIsNone(result.exception, result.output)
        for key, value in profile.items():
            self.assertIn(key.lower(), result.output)
            self.assertIn(value, result.output)

    def test_delete(self):
        """
        Test for verdi profile delete command
        """
        from aiida.cmdline.commands.cmd_profile import profile_delete, profile_list

        # delete single profile
        result = self.runner.invoke(profile_delete, ["--force", self.dummy_profile_list[2]])
        self.assertIsNone(result.exception)

        result = self.runner.invoke(profile_list)
        self.assertIsNone(result.exception)

        self.assertNotIn(self.dummy_profile_list[2], result.output)
        self.assertIsNone(result.exception)

        # delete multiple profile
        result = self.runner.invoke(profile_delete, ["--force", self.dummy_profile_list[3], self.dummy_profile_list[4]])
        self.assertIsNone(result.exception)

        result = self.runner.invoke(profile_list)
        self.assertIsNone(result.exception)
        self.assertNotIn(self.dummy_profile_list[3], result.output)
        self.assertNotIn(self.dummy_profile_list[4], result.output)
        self.assertIsNone(result.exception)


def get_random_string(length):
    import string
    import random

    return ''.join(random.choice(string.ascii_letters) for m in range(length))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-arguments
"""`verdi work` command."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.cmdline.utils.query.calculation import CalculationQueryBuilder
from aiida.common.log import LOG_LEVELS


@verdi.group('work')
def verdi_work():
    """Inspect and manage work calculations."""
    pass


@verdi_work.command('list')
@options.PROJECT(
    type=click.Choice(CalculationQueryBuilder.valid_projections), default=CalculationQueryBuilder.default_projections)
@options.ALL(help='Show all entries, regardless of their process state.')
@options.PROCESS_STATE()
@options.EXIT_STATUS()
@options.FAILED()
@options.PAST_DAYS()
@options.LIMIT()
@options.RAW()
@decorators.with_dbenv()
def work_list(all_entries, process_state, exit_status, failed, past_days, limit, project, raw):
    """Show a list of work calculations that are still running."""
    from tabulate import tabulate
    from aiida.cmdline.utils.common import print_last_process_state_change
    from aiida.orm.calculation.function import FunctionCalculation
    from aiida.orm.calculation.work import WorkCalculation

    node_types = (WorkCalculation, FunctionCalculation)

    builder = CalculationQueryBuilder()
    filters = builder.get_filters(all_entries, process_state, exit_status, failed, node_types=node_types)
    query_set = builder.get_query_set(filters=filters, past_days=past_days, limit=limit)
    projected = builder.get_projected(query_set, projections=project)

    headers = projected.pop(0)

    if raw:
        tabulated = tabulate(projected, tablefmt='plain')
        echo.echo(tabulated)
    else:
        tabulated = tabulate(projected, headers=headers)
        echo.echo(tabulated)
        echo.echo('\nTotal results: {}\n'.format(len(projected)))
        print_last_process_state_change(process_type='work')


@verdi_work.command('report')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@click.option('-i', '--indent-size', type=int, default=2, help='set the number of spaces to indent each level by')
@click.option(
    '-l',
    '--levelname',
    type=click.Choice(LOG_LEVELS.keys()),
    default='REPORT',
    help='filter the results by name of the log level')
@click.option('-m', '--max-depth', 'max_depth', type=int, default=None, help='limit the number of levels to be printed')
def work_report(calculations, levelname, indent_size, max_depth):
    # pylint: disable=too-many-locals
    """
    Return a list of recorded log messages for the WorkChain with pk=PK
    """
    import itertools
    from aiida.orm.backend import construct_backend
    from aiida.orm.querybuilder import QueryBuilder
    from aiida.orm.calculation.work import WorkCalculation

    def get_report_messages(pk, depth, levelname):
        """Return list of log messages with given levelname and their depth for a node with a given pk."""
        backend = construct_backend()
        filters = {
            'objpk': pk,
        }

        entries = backend.logs.find(filter_by=filters)
        entries = [entry for entry in entries if LOG_LEVELS[entry.levelname] >= LOG_LEVELS[levelname]]
        return [(_, depth) for _ in entries]

    def get_subtree(pk, level=0):
        """Get a nested tree of work calculation nodes and their nesting level starting from this pk."""
        builder = QueryBuilder()
        builder.append(cls=WorkCalculation, filters={'id': pk}, tag='workcalculation')
        builder.append(
            cls=WorkCalculation,
            project=['id'],
            # In the future, we should specify here the type of link
            # for now, CALL links are the only ones allowing calc-calc
            # (we here really want instead to follow CALL links)
            output_of='workcalculation',
            tag='subworkchains')
        result = list(itertools.chain(*builder.distinct().all()))

        # This will return a single flat list of tuples, where the first element
        # corresponds to the WorkChain pk and the second element is an integer
        # that represents its level of nesting within the chain
        return [(pk, level)] + list(itertools.chain(*[get_subtree(subpk, level=level + 1) for subpk in result]))

    def print_subtree(tree, prepend=""):
        echo.echo("{}{}".format(prepend, tree[0]))
        for subtree in tree[1]:
            print_subtree(subtree, prepend=prepend + "  ")

    for calculation in calculations:

        workchain_tree = get_subtree(calculation.pk)

        if max_depth:
            report_list = [
                get_report_messages(pk, depth, levelname) for pk, depth in workchain_tree if depth < max_depth
            ]
        else:
            report_list = [get_report_messages(pk, depth, levelname) for pk, depth in workchain_tree]

        reports = list(itertools.chain(*report_list))
        reports.sort(key=lambda r: r[0].time)

        if not reports:
            echo.echo("No log messages recorded for this work calculation")
            return

        object_ids = [entry[0].id for entry in reports]
        levelnames = [len(entry[0].levelname) for entry in reports]
        width_id = len(str(max(object_ids)))
        width_levelname = max(levelnames)
        for entry, depth in reports:
            echo.echo('{time:%Y-%m-%d %H:%M:%S} [{id:<{width_id}} | {levelname:>{width_levelname}}]:{indent} {message}'.
                      format(
                          id=entry.id,
                          levelname=entry.levelname,
                          message=entry.message,
                          time=entry.time,
                          width_id=width_id,
                          width_levelname=width_levelname,
                          indent=' ' * (depth * indent_size)))

    return


@verdi_work.command('show')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@decorators.with_dbenv()
def work_show(calculations):
    """Show a summary for one or multiple calculations."""
    from aiida.cmdline.utils.common import get_node_info

    for calculation in calculations:
        echo.echo(get_node_info(calculation))


@verdi_work.command('status')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
def work_status(calculations):
    """
    Print the status of work calculations
    """
    from aiida.utils.ascii_vis import format_call_graph

    for calculation in calculations:
        graph = format_call_graph(calculation)
        echo.echo(graph)


@verdi_work.command('plugins')
@click.argument('entry_point', type=click.STRING, required=False)
@decorators.with_dbenv()
def work_plugins(entry_point):
    """
    Print a list of registered workflow plugins or details of a specific workflow plugin
    """
    from aiida.common.exceptions import LoadingPluginFailed, MissingPluginError
    from aiida.plugins.entry_point import get_entry_point_names, load_entry_point

    if entry_point:
        try:
            plugin = load_entry_point('aiida.workflows', entry_point)
        except (LoadingPluginFailed, MissingPluginError) as exception:
            echo.echo_critical(exception)
        else:
            echo.echo_info(entry_point)
            echo.echo(plugin.get_description())
    else:
        entry_points = get_entry_point_names('aiida.workflows')
        if entry_points:
            echo.echo('Registered workflow entry points:')
            for registered_entry_point in entry_points:
                echo.echo("* {}".format(registered_entry_point))

            echo.echo('')
            echo.echo_info('Pass the entry point as an argument to display detailed information')
        else:
            echo.echo_error('No workflow plugins found')


@verdi_work.command('kill')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@decorators.deprecated_command("This command will be removed in a future release. Use 'verdi process kill' instead.")
def work_kill(calculations):
    """
    Kill work calculations
    """
    from aiida.work import RemoteException, DeliveryFailed, new_blocking_control_panel

    with new_blocking_control_panel() as control_panel:
        for calculation in calculations:

            if calculation.is_terminated:
                echo.echo_error('Calculation<{}> is already terminated'.format(calculation.pk))
                continue

            try:
                if control_panel.kill_process(calculation.pk):
                    echo.echo_success('killed Calculation<{}>'.format(calculation.pk))
                else:
                    echo.echo_error('problem killing Calculation<{}>'.format(calculation.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_error('failed to kill Calculation<{}>: {}'.format(calculation.pk, exception))


@verdi_work.command('pause')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@decorators.deprecated_command("This command will be removed in a future release. Use 'verdi process pause' instead.")
def work_pause(calculations):
    """
    Pause running work calculations
    """
    from aiida.work import RemoteException, DeliveryFailed, new_blocking_control_panel

    with new_blocking_control_panel() as control_panel:
        for calculation in calculations:

            if calculation.is_terminated:
                echo.echo_error('Calculation<{}> is already terminated'.format(calculation.pk))
                continue

            try:
                if control_panel.pause_process(calculation.pk):
                    echo.echo_success('paused Calculation<{}>'.format(calculation.pk))
                else:
                    echo.echo_error('problem pausing Calculation<{}>'.format(calculation.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_error('failed to pause Calculation<{}>: {}'.format(calculation.pk, exception))


@verdi_work.command('play')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@decorators.deprecated_command("This command will be removed in a future release. Use 'verdi process play' instead.")
def work_play(calculations):
    """
    Play paused work calculations
    """
    from aiida.work import RemoteException, DeliveryFailed, new_blocking_control_panel

    with new_blocking_control_panel() as control_panel:
        for calculation in calculations:

            if calculation.is_terminated:
                echo.echo_error('Calculation<{}> is already terminated'.format(calculation.pk))
                continue

            try:
                if control_panel.play_process(calculation.pk):
                    echo.echo_success('played Calculation<{}>'.format(calculation.pk))
                else:
                    echo.echo_critical('problem playing Calculation<{}>'.format(calculation.pk))
            except (RemoteException, DeliveryFailed) as exception:
                echo.echo_critical('failed to play Calculation<{}>: {}'.format(calculation.pk, exception))


@verdi_work.command('watch')
@arguments.CALCULATIONS(
    type=types.CalculationParamType(sub_classes=('aiida.calculations:work', 'aiida.calculations:function')))
@decorators.deprecated_command("This command will be removed in a future release. Use 'verdi process watch' instead.")
def work_watch(calculations):
    """
    Watch the state transitions for work calculations
    """
    from kiwipy import BroadcastFilter
    from aiida.work.rmq import create_communicator

    def _print(body, sender, subject, correlation_id):
        echo.echo("pk={}, subject={}, body={}, correlation_id={}".format(sender, subject, body, correlation_id))

    communicator = create_communicator()

    for calculation in calculations:

        if calculation.is_terminated:
            echo.echo_warning('Calculation<{}> is already terminated'.format(calculation.pk))
        communicator.add_broadcast_subscriber(BroadcastFilter(_print, sender=calculation.pk))

    try:
        communicator.await()
    except (SystemExit, KeyboardInterrupt):
        try:
            communicator.disconnect()
        except RuntimeError:
            pass



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
This allows to manage profiles from command line.
"""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.utils import echo
from aiida.cmdline.params import options, types
from aiida.control.postgres import Postgres
from aiida.common.setup import get_default_profile_name


@verdi.group('profile')
def verdi_profile():
    """Inspect and manage the configured profiles."""
    pass


@verdi_profile.command('list')
def profile_list():
    """Displays list of all available profiles."""
    from aiida.common.setup import get_profiles_list, AIIDA_CONFIG_FOLDER
    from aiida.common.exceptions import ConfigurationError

    echo.echo_info('configuration folder: {}'.format(AIIDA_CONFIG_FOLDER))

    try:
        default_profile = get_default_profile_name()
    except ConfigurationError as exception:
        echo.echo_critical('could not load the configuration file: {}'.format(exception))

    if default_profile is None:
        echo.echo_critical('no default profile configured yet, run `verdi setup`')
    else:
        echo.echo_info('default profile is highlighted and marked by the * symbol')

    for profile in sorted(get_profiles_list()):
        if profile == default_profile:
            color = 'green'
            symbol = '*'
        else:
            color = ''
            symbol = ' '

        click.secho('{} {}'.format(symbol, profile), fg=color)


@verdi_profile.command('show')
@click.argument('profile', nargs=1, type=types.ProfileParamType(), default=get_default_profile_name())
def profile_show(profile):
    """Show details for PROFILE or, when not specified, the default profile."""
    import tabulate

    headers = ('Attribute', 'Value')
    data = sorted([(k.lower(), v) for k, v in profile.config.items()])
    echo.echo(tabulate.tabulate(data, headers=headers))


@verdi_profile.command('setdefault')
@click.argument('profile', nargs=1, type=types.ProfileParamType())
def profile_setdefault(profile):
    """Set PROFILE as the default profile."""
    from aiida.common.setup import set_default_profile
    set_default_profile(profile.name, force_rewrite=True)
    echo.echo_success('{} set as default profile'.format(profile.name))


@verdi_profile.command('delete')
@options.FORCE(help='to skip any questions/warnings about loss of data')
@click.argument('profiles', nargs=-1, type=types.ProfileParamType())
def profile_delete(force, profiles):
    """
    Delete PROFILES separated by space from aiida config file
    along with its associated database and repository.
    """
    from aiida.common.setup import get_or_create_config, update_config
    import os.path
    from six.moves.urllib.parse import urlparse  # pylint: disable=import-error

    profile_names = [profile.name for profile in profiles]

    echo.echo('profiles: {}'.format(', '.join(profile_names)))

    confs = get_or_create_config()
    available_profiles = confs.get('profiles', {})
    users = [available_profiles[name].get('AIIDADB_USER', '') for name in available_profiles.keys()]

    for profile_name in profile_names:
        try:
            profile = available_profiles[profile_name]
        except KeyError:
            echo.echo_info("Profile '{}' does not exist".format(profile_name))
            continue

        postgres = Postgres(port=profile.get('AIIDADB_PORT'), interactive=True, quiet=False)
        postgres.dbinfo["user"] = profile.get('AIIDADB_USER')
        postgres.dbinfo["host"] = profile.get('AIIDADB_HOST')
        postgres.determine_setup()

        import aiida.utils.json as json
        echo.echo(json.dumps(postgres.dbinfo, indent=4))

        db_name = profile.get('AIIDADB_NAME', '')
        if not postgres.db_exists(db_name):
            echo.echo_info("Associated database '{}' does not exist.".format(db_name))
        elif force or click.confirm("Delete associated database '{}'?\n"
                                    "WARNING: All data will be lost.".format(db_name)):
            echo.echo_info("Deleting database '{}'.".format(db_name))
            postgres.drop_db(db_name)

        user = profile.get('AIIDADB_USER', '')
        if not postgres.dbuser_exists(user):
            echo.echo_info("Associated database user '{}' does not exist.".format(user))
        elif users.count(user) > 1:
            echo.echo_info("Associated database user '{}' is used by other profiles "
                           "and will not be deleted.".format(user))
        elif force or click.confirm("Delete database user '{}'?".format(user)):
            echo.echo_info("Deleting user '{}'.".format(user))
            postgres.drop_dbuser(user)

        repo_uri = profile.get('AIIDADB_REPOSITORY_URI', '')
        repo_path = urlparse(repo_uri).path
        repo_path = os.path.expanduser(repo_path)
        if not os.path.isabs(repo_path):
            echo.echo_info("Associated file repository '{}' does not exist.".format(repo_path))
        elif not os.path.isdir(repo_path):
            echo.echo_info("Associated file repository '{}' is not a directory.".format(repo_path))
        elif force or click.confirm("Delete associated file repository '{}'?\n"
                                    "WARNING: All data will be lost.".format(repo_path)):
            echo.echo_info("Deleting directory '{}'.".format(repo_path))
            import shutil
            shutil.rmtree(repo_path)

        if force or click.confirm(
                "Delete configuration for profile '{}'?\n"
                "WARNING: Permanently removes profile from the list of AiiDA profiles.".format(profile_name)):
            echo.echo_info("Deleting configuration for profile '{}'.".format(profile_name))
            del available_profiles[profile_name]
            update_config(confs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
The dummy model encodes the model defined by django in backends.djsite
using SQLAlchemy.
This is done to query the database with more performant ORM of SA.
"""

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (
    Column, Table, ForeignKey, UniqueConstraint, create_engine,
    select, func, join, and_, or_, not_, except_, case, exists,
    text
)

from sqlalchemy.types import (
    Integer, String, DateTime, Float, Boolean, Text,
)
from sqlalchemy.orm import (
    relationship,
    backref,
    sessionmaker,
    foreign, mapper, aliased
)

from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import Cast
from sqlalchemy.dialects.postgresql import UUID, JSONB, INTEGER, array
# TO COMPILE MY OWN FUNCTIONALITIES:
from sqlalchemy.sql.expression import FunctionElement, cast
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.ext.compiler import compiles

# SETTINGS:
from aiida.common.setup import get_profile_config
from aiida.backends import settings

# EXCEPTIONS
from aiida.common.exceptions import DbContentError, MissingPluginError

# MISC
from aiida.backends.sqlalchemy.models.utils import uuid_func
from aiida.utils import timezone
from aiida.common.datastructures import (calc_states, _sorted_datastates,
                                         sort_states)

Base = declarative_base()


class DbLink(Base):
    __tablename__ = "db_dblink"
    id = Column(Integer, primary_key=True)
    input_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    )
    output_id = Column(
        Integer,
        ForeignKey(
            'db_dbnode.id',
            ondelete="CASCADE",
            deferrable=True,
            initially="DEFERRED"
        )
    )
    type = Column(String(255))
    input = relationship("DbNode", primaryjoin="DbLink.input_id == DbNode.id")
    output = relationship("DbNode", primaryjoin="DbLink.output_id == DbNode.id")
    label = Column(String(255), index=True, nullable=False)


class DbCalcState(Base):
    __tablename__ = "db_dbcalcstate"
    id = Column(Integer, primary_key=True)
    dbnode = relationship(
        'DbNode',
        backref=backref('dbstates', passive_deletes=True)
    )
    state = Column(String(255))
    time = Column(DateTime(timezone=True), default=timezone.now)
    dbnode_id = Column(
        Integer,
        ForeignKey(
            'db_dbnode.id', ondelete="CASCADE",
            deferrable=True, initially="DEFERRED"
        )
    )
    __table_args__ = (
        UniqueConstraint('dbnode_id', 'state'),
    )


class DbAttribute(Base):
    __tablename__ = "db_dbattribute"
    id = Column(Integer, primary_key=True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable=True)


class DbExtra(Base):
    __tablename__ = "db_dbextra"
    id = Column(Integer, primary_key=True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable=True)


class DbComputer(Base):
    __tablename__ = "db_dbcomputer"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255))

    description = Column(Text, nullable=True)
    enabled = Column(Boolean)

    transport_type = Column(String(255))
    scheduler_type = Column(String(255))

    transport_params = Column(String(255))
    _metadata = Column('metadata', String(255), default="{}")

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbComputer as DjangoSchemaDbComputer
        djcomputer = DjangoSchemaDbComputer(
            id=self.id, uuid=self.uuid, name=self.name,
            hostname=self.hostname, description=self.description,
            enabled=self.enabled, transport_type=self.transport_type,
            scheduler_type=self.scheduler_type, transport_params=self.transport_params,
            metadata=self._metadata
        )
        return djcomputer.get_aiida_class()


class DbUser(Base):
    __tablename__ = "db_dbuser"

    id = Column(Integer, primary_key=True)
    email = Column(String(254), unique=True, index=True)
    password = Column(String(128))  # Clear text password ?
    first_name = Column(String(254), nullable=True)
    last_name = Column(String(254), nullable=True)
    institution = Column(String(254), nullable=True)

    is_staff = Column(Boolean, default=False)
    is_active = Column(Boolean, default=False)

    last_login = Column(DateTime(timezone=True), default=timezone.now)
    date_joined = Column(DateTime(timezone=True), default=timezone.now)

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbUser as DjangoSchemaDbUser
        djuser = DjangoSchemaDbUser(
            id=self.id, email=self.email, password=self.password,
            first_name=self.first_name, last_name=self.last_name,
            institution=self.institution, is_staff=self.is_staff,
            is_active=self.is_active, last_login=self.last_login,
            date_joined=self.date_joined
        )
        return djuser.get_aiida_class()


table_groups_nodes = Table(
    'db_dbgroup_dbnodes',
    Base.metadata,
    Column('id', Integer, primary_key=True),
    Column(
        'dbnode_id', Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    ),
    Column(
        'dbgroup_id', Integer,
        ForeignKey('db_dbgroup.id', deferrable=True, initially="DEFERRED")
    )
)


class DbGroup(Base):
    __tablename__ = "db_dbgroup"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), index=True)

    type = Column(String(255), default="", index=True)

    time = Column(DateTime(timezone=True), default=timezone.now)
    description = Column(Text, nullable=True)

    user_id = Column(
        Integer,
        ForeignKey(
            'db_dbuser.id', ondelete='CASCADE',
            deferrable=True, initially="DEFERRED")
    )
    user = relationship('DbUser', backref=backref('dbgroups', cascade='merge'))

    dbnodes = relationship('DbNode', secondary=table_groups_nodes,
                           backref="dbgroups", lazy='dynamic')

    __table_args__ = (
        UniqueConstraint('name', 'type'),
    )

    def __str__(self):
        if self.type:
            return '<DbGroup [type: {}] "{}">'.format(self.type, self.name)
        else:
            return '<DbGroup [user-defined] "{}">'.format(self.name)

    def get_aiida_class(self):
        from aiida.orm.implementation.django.group import Group as DjangoAiidaGroup
        from aiida.backends.djsite.db.models import DbGroup as DjangoSchemaDbGroup
        dbgroup = DjangoSchemaDbGroup(
            id=self.id,
            type=self.type,
            uuid=self.uuid,
            name=self.name,
            time=self.time,
            description=self.description,
            user_id=self.user_id,
        )

        return DjangoAiidaGroup(dbgroup=dbgroup)


class DbNode(Base):
    __tablename__ = "db_dbnode"
    id = Column(Integer, primary_key=True)
    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    type = Column(String(255), index=True)
    process_type = Column(String(255), index=True)
    label = Column(String(255), index=True, nullable=True)
    description = Column(Text(), nullable=True)
    ctime = Column(DateTime(timezone=True), default=timezone.now)
    mtime = Column(DateTime(timezone=True), default=timezone.now)
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially="DEFERRED"),
        nullable=True
    )
    dbcomputer = relationship(
        'DbComputer',
        backref=backref('dbnodes', passive_deletes=True)
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially="DEFERRED"),
        nullable=False
    )
    user = relationship('DbUser', backref='dbnodes')

    public = Column(Boolean, default=False)

    nodeversion = Column(Integer, default=1)

    attributes = relationship('DbAttribute', uselist=True, backref='dbnode')
    extras = relationship('DbExtra', uselist=True, backref='dbnode')

    outputs = relationship(
        "DbNode",
        secondary="db_dblink",
        primaryjoin="DbNode.id == DbLink.input_id",
        secondaryjoin="DbNode.id == DbLink.output_id",
        backref=backref("inputs", passive_deletes=True),
        passive_deletes=True
    )

    def get_aiida_class(self):
        """
        Return the corresponding instance of
        :func:`~aiida.orm.implementation.django.node.Node`
        or a subclass return by the plugin loader.

        .. todo::
            The behavior is quite pathetic, creating a django DbNode instance
            to instantiate the aiida instance.
            These means that every time you load Aiida instances with
            the QueryBuilder when using Django as a backend, three instances
            are instantiated for every Aiida instance you load!
            Could be fixed by allowing DbNode from the dummy nodel to be passed
            to AiidaNode's __init__.

        :returns: An instance of the plugin class
        """
        # I need to import the DbNode in the Django model,
        # and instantiate an object that has the same attributes as self.
        from aiida.backends.djsite.db.models import DbNode as DjangoSchemaDbNode
        dbnode = DjangoSchemaDbNode(
            id=self.id, type=self.type, process_type=self.process_type, uuid=self.uuid, ctime=self.ctime,
            mtime=self.mtime, label=self.label, description=self.description, dbcomputer_id=self.dbcomputer_id,
            user_id=self.user_id, public=self.public, nodeversion=self.nodeversion
        )
        return dbnode.get_aiida_class()

    @hybrid_property
    def user_email(self):
        """
        Returns: the email of the user
        """
        return self.user.email

    @user_email.expression
    def user_email(cls):
        """
        Returns: the email of the user at a class level (i.e. in the database)
        """
        return select([DbUser.email]).where(DbUser.id == cls.user_id).label(
            'user_email')

    # Computer name
    @hybrid_property
    def computer_name(self):
        """
        Returns: the of the computer
        """
        return self.dbcomputer.name

    @computer_name.expression
    def computer_name(cls):
        """
        Returns: the name of the computer at a class level (i.e. in the 
        database)
        """
        return select([DbComputer.name]).where(DbComputer.id ==
                                               cls.dbcomputer_id).label(
            'computer_name')

    # State
    @hybrid_property
    def state(self):
        """
        Return the most recent state from DbCalcState
        """
        if not self.id:
            return None
        all_states = DbCalcState.query.filter(DbCalcState.dbnode_id == self.id).all()
        if all_states:
            # return max((st.time, st.state) for st in all_states)[1]
            return sort_states(((dbcalcstate.state, dbcalcstate.state.value)
                                for dbcalcstate in all_states),
                               use_key=True)[0]
        else:
            return None

    @state.expression
    def state(cls):
        """
        Return the expression to get the 'latest' state from DbCalcState,
        to be used in queries, where 'latest' is defined using the state order
        defined in _sorted_datastates.
        """
        # Sort first the latest states
        whens = {
            v: idx for idx, v
            in enumerate(_sorted_datastates[::-1], start=1)}
        custom_sort_order = case(value=DbCalcState.state,
                                 whens=whens,
                                 else_=100)  # else: high value to put it at the bottom

        # Add numerical state to string, to allow to sort them
        states_with_num = select([
            DbCalcState.id.label('id'),
            DbCalcState.dbnode_id.label('dbnode_id'),
            DbCalcState.state.label('state_string'),
            custom_sort_order.label('num_state')
        ]).select_from(DbCalcState).alias()

        # Get the most 'recent' state (using the state ordering, and the min function) for
        # each calc
        calc_state_num = select([
            states_with_num.c.dbnode_id.label('dbnode_id'),
            func.min(states_with_num.c.num_state).label('recent_state')
        ]).group_by(states_with_num.c.dbnode_id).alias()

        # Join the most-recent-state table with the DbCalcState table
        all_states_q = select([
            DbCalcState.dbnode_id.label('dbnode_id'),
            DbCalcState.state.label('state_string'),
            calc_state_num.c.recent_state.label('recent_state'),
            custom_sort_order.label('num_state'),
        ]).select_from(  # DbCalcState).alias().join(
            join(DbCalcState, calc_state_num, DbCalcState.dbnode_id == calc_state_num.c.dbnode_id)).alias()

        # Get the association between each calc and only its corresponding most-recent-state row
        subq = select([
            all_states_q.c.dbnode_id.label('dbnode_id'),
            all_states_q.c.state_string.label('state')
        ]).select_from(all_states_q).where(all_states_q.c.num_state == all_states_q.c.recent_state).alias()

        # Final filtering for the actual query
        return select([subq.c.state]). \
            where(
            subq.c.dbnode_id == cls.id,
        ). \
            label('laststate')


profile = get_profile_config(settings.AIIDADB_PROFILE)


def get_aldjemy_session():
    """
    Use aldjemy to make a session

    .. note:
        Use only in this case. In normal production mode
        it is safer make session explictly because it is more robust
    """
    from aldjemy.core import get_engine
    engine = get_engine()
    _Session = sessionmaker(bind=engine)
    return _Session()


session = get_aldjemy_session()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Plugin aware click command Group."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.plugins.entry_point import load_entry_point, get_entry_point_names, MissingEntryPointError


class Pluginable(click.Group):
    """A click command group that finds and loads plugin commands lazily."""

    def __init__(self, *args, **kwargs):
        """Initialize with entry point group."""
        self._entry_point_group = kwargs.pop('entry_point_group')
        super(Pluginable, self).__init__(*args, **kwargs)

    def list_commands(self, ctx):
        """Add entry point names of available plugins to the command list."""
        subcommands = super(Pluginable, self).list_commands(ctx)
        subcommands.extend(get_entry_point_names(self._entry_point_group))
        return subcommands

    def get_command(self, ctx, name):  # pylint: disable=arguments-differ
        """Try to load a subcommand from entry points, else defer to super."""
        command = None
        try:
            command = load_entry_point(self._entry_point_group, name)
        except MissingEntryPointError:
            command = super(Pluginable, self).get_command(ctx, name)
        return command



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from functools import reduce

from aiida.backends.djsite.db.models import DbUser
from aiida.orm.user import User, UserCollection
from aiida.utils.email import normalize_email
from aiida.common.utils import type_check
from . import utils


class DjangoUserCollection(UserCollection):

    def create(self, email, first_name='', last_name='', institution=''):
        """
        Create a user with the provided email address

        :return: A new user object
        :rtype: :class:`aiida.orm.User`
        """
        return DjangoUser(self.backend,
                          email=normalize_email(email),
                          first_name=first_name,
                          last_name=last_name,
                          institution=institution)

    def find(self, email=None, id=None):
        # Constructing the default query
        import operator
        from django.db.models import Q
        query_list = []

        # If an id is specified then we add it to the query
        if id is not None:
            query_list.append(Q(pk=id))

        # If an email is specified then we add it to the query
        if email is not None:
            query_list.append(Q(email=email))

        if not query_list:
            dbusers = DbUser.objects.all()
        else:
            dbusers = DbUser.objects.filter(reduce(operator.and_, query_list))
        users = []
        for dbuser in dbusers:
            users.append(self.from_dbmodel(dbuser))
        return users

    def from_dbmodel(self, dbmodel):
        return DjangoUser.from_dbmodel(dbmodel, self.backend)


class DjangoUser(User):

    @classmethod
    def from_dbmodel(cls, dbmodel, backend):
        """
        Create a DjangoUser from a dbmodel instance

        :param backend: The backend
        :type backend: :class:`DjangoUserCollection`
        :param dbmodel: The dbmodel instance
        :type dbmodel: :class:`aiida.backends.djsite.db.models.DbUser`
        :return: A DjangoUser instance
        :rtype: :class:`DjangoUser`
        """
        type_check(dbmodel, DbUser)
        user = cls.__new__(cls)
        super(DjangoUser, user).__init__(backend)
        user._dbuser = utils.ModelWrapper(dbmodel)
        return user

    def __init__(self, backend, email, first_name, last_name, institution):
        super(DjangoUser, self).__init__(backend)
        self._dbuser = utils.ModelWrapper(DbUser(
            email=email,
            first_name=first_name,
            last_name=last_name,
            institution=institution))

    @property
    def dbuser(self):
        # We have to get the underlying model here rather than just the wrapper
        return self._dbuser._model

    @property
    def pk(self):
        return self._dbuser.pk

    @property
    def id(self):
        return self._dbuser.pk

    def __int__(self):
        # Needed to pass this object to raw django queries
        return self.id

    @property
    def is_stored(self):
        return self._dbuser.pk is not None

    def store(self):
        self._dbuser.save()

    @property
    def email(self):
        return self._dbuser.email

    @email.setter
    def email(self, val):
        self._dbuser.email = val

    def _set_password(self, val):
        self._dbuser.password = val

    def _get_password(self):
        return self._dbuser.password

    @property
    def first_name(self):
        return self._dbuser.first_name

    @first_name.setter
    def first_name(self, val):
        self._dbuser.first_name = val

    @property
    def last_name(self):
        return self._dbuser.last_name

    @last_name.setter
    def last_name(self, val):
        self._dbuser.last_name = val

    @property
    def institution(self):
        return self._dbuser.institution

    @institution.setter
    def institution(self, val):
        self._dbuser.institution = val

    @property
    def is_active(self):
        return self._dbuser.is_active

    @is_active.setter
    def is_active(self, val):
        self._dbuser.is_active = val

    @property
    def last_login(self):
        return self._dbuser.last_login

    @last_login.setter
    def last_login(self, val):
        self._dbuser.last_login = val

    @property
    def date_joined(self):
        return self._dbuser.date_joined

    @date_joined.setter
    def date_joined(self, val):
        self._dbuser.date_joined = val



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

from six.moves import range
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase

COMMENT = u'Well I never...'


class TestVerdiUserCommand(AiidaTestCase):

    def setUp(self):
        self.cli_runner = CliRunner()

    def test_comment_show_simple(self):
        """ test simply calling the show command (without data to show) """
        from aiida.cmdline.commands.cmd_comment import show

        result = CliRunner().invoke(show, [], catch_exceptions=False)
        self.assertEqual(result.output, "")
        self.assertEqual(result.exit_code, 0)

    def test_comment_show(self):
        """ Test showing an existing comment """
        from aiida.cmdline.commands.cmd_comment import show
        from aiida.orm import Node

        node = Node()
        node.store()
        node.add_comment(COMMENT)

        result = CliRunner().invoke(show, [str(node.pk)], catch_exceptions=False)
        self.assertNotEqual(result.output.find(COMMENT), -1)
        self.assertEqual(result.exit_code, 0)

    def test_comment_add(self):
        """ Test adding a comment """
        from aiida.cmdline.commands.cmd_comment import add
        from aiida.orm import Node

        node = Node()
        node.store()

        result = CliRunner().invoke(add, ['-c{}'.format(COMMENT), str(node.pk)], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

        comment = node.get_comments()
        self.assertEquals(len(comment), 1)
        self.assertEqual(comment[0]['content'], COMMENT)

    def test_comment_remove(self):
        """ Test removing a comment """
        from aiida.cmdline.commands.cmd_comment import remove
        from aiida.orm import Node

        node = Node()
        node.store()
        comment_id = node.add_comment(COMMENT)

        self.assertEquals(len(node.get_comments()), 1)

        result = CliRunner().invoke(remove, [str(node.pk), str(comment_id), '--force'], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

        self.assertEquals(len(node.get_comments()), 0)

    def test_comment_remove_all(self):
        """ Test removing all comments from a node """
        from aiida.cmdline.commands.cmd_comment import remove
        from aiida.orm import Node

        node = Node()
        node.store()
        for _ in range(10):
            node.add_comment(COMMENT)

        self.assertEqual(len(node.get_comments()), 10)

        result = CliRunner().invoke(remove, [str(node.pk), '--all', '--force'], catch_exceptions=False)
        self.assertEqual(result.exit_code, 0)

        self.assertEqual(len(node.get_comments()), 0)




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=protected-access
"""Tests for `verdi work`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_work
from aiida.common.links import LinkType
from aiida.common.log import LOG_LEVEL_REPORT
from aiida.orm.calculation.function import FunctionCalculation
from aiida.orm.calculation.work import WorkCalculation
from aiida.work import runners, rmq, test_utils


def get_result_lines(result):
    return [e for e in result.output.split('\n') if e]


class TestVerdiWork(AiidaTestCase):
    """Tests for `verdi work`."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiWork, cls).setUpClass(*args, **kwargs)
        rmq_config = rmq.get_rmq_config()

        # These two need to share a common event loop otherwise the first will never send
        # the message while the daemon is running listening to intercept
        cls.runner = runners.Runner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)

        cls.daemon_runner = runners.DaemonRunner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)

    def setUp(self):
        super(TestVerdiWork, self).setUp()
        self.cli_runner = CliRunner()

    def test_status(self):
        """Test the status command."""
        calc = self.runner.submit(test_utils.WaitProcess)
        result = self.cli_runner.invoke(cmd_work.work_status, [str(calc.pk)])
        self.assertIsNone(result.exception)

    def test_list(self):
        """Test the list command."""
        from aiida.work.processes import ProcessState

        # Number of output lines in -r/--raw format should be zero when there are no calculations yet
        result = self.cli_runner.invoke(cmd_work.work_list, ['-r'])
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 0)

        calcs = []

        # Create 6 FunctionCalculations and WorkCalculations (one for each ProcessState)
        for state in ProcessState:

            calc = FunctionCalculation()
            calc._set_process_state(state)

            # Set the FunctionCalculation as successful
            if state == ProcessState.FINISHED:
                calc._set_exit_status(0)

            calc.store()
            calcs.append(calc)

            calc = WorkCalculation()
            calc._set_process_state(state)

            # Set the WorkCalculation as failed
            if state == ProcessState.FINISHED:
                calc._set_exit_status(1)

            calc.store()
            calcs.append(calc)

        # Default behavior should yield all active states (CREATED, RUNNING and WAITING) so six in total
        result = self.cli_runner.invoke(cmd_work.work_list, ['-r'])
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 6)

        # Adding the all option should return all entries regardless of process state
        for flag in ['-a', '--all']:
            result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag])
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 12)

        # Passing the limit option should limit the results
        for flag in ['-l', '--limit']:
            result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag, '6'])
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 6)

        # Filtering for a specific process state
        for flag in ['-S', '--process-state']:
            for flag_value in ['created', 'running', 'waiting', 'killed', 'excepted', 'finished']:
                result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag, flag_value])
                self.assertIsNone(result.exception)
                self.assertEquals(len(get_result_lines(result)), 2)

        # Filtering for exit status should only get us one
        for flag in ['-E', '--exit-status']:
            for exit_status in ['0', '1']:
                result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag, exit_status])
                self.assertIsNone(result.exception)
                self.assertEquals(len(get_result_lines(result)), 1)

        # Passing the failed flag as a shortcut for FINISHED + non-zero exit status
        for flag in ['-X', '--failed']:
            result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag])
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 1)

        # Projecting on pk should allow us to verify all the pks
        for flag in ['-P', '--project']:
            result = self.cli_runner.invoke(cmd_work.work_list, ['-r', flag, 'pk'])
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 6)

            for line in get_result_lines(result):
                self.assertIn(line.strip(), [str(calc.pk) for calc in calcs])

    def test_report(self):
        """Test the report command."""
        grandparent = WorkCalculation().store()
        parent = WorkCalculation().store()
        child = WorkCalculation().store()

        parent.add_link_from(grandparent, link_type=LinkType.CALL)
        child.add_link_from(parent, link_type=LinkType.CALL)

        grandparent.logger.log(LOG_LEVEL_REPORT, 'grandparent_message')
        parent.logger.log(LOG_LEVEL_REPORT, 'parent_message')
        child.logger.log(LOG_LEVEL_REPORT, 'child_message')

        result = self.cli_runner.invoke(cmd_work.work_report, [str(grandparent.pk)])
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 3)

        result = self.cli_runner.invoke(cmd_work.work_report, [str(parent.pk)])
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 2)

        result = self.cli_runner.invoke(cmd_work.work_report, [str(child.pk)])
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 1)

        # Max depth should limit nesting level
        for flag in ['-m', '--max-depth']:
            for flag_value in [1, 2]:
                result = self.cli_runner.invoke(cmd_work.work_report, [str(grandparent.pk), flag, str(flag_value)])
                self.assertIsNone(result.exception)
                self.assertEquals(len(get_result_lines(result)), flag_value)

        # Filtering for other level name such as WARNING should not have any hits and only print the no log message
        for flag in ['-l', '--levelname']:
            result = self.cli_runner.invoke(cmd_work.work_report, [str(grandparent.pk), flag, 'WARNING'])
            self.assertIsNone(result.exception)
            self.assertEquals(len(get_result_lines(result)), 1)
            self.assertEquals(get_result_lines(result)[0], 'No log messages recorded for this work calculation')

    def test_plugins(self):
        """Test the plugins command. As of writing there are no default plugins defined for aiida-core."""
        result = self.cli_runner.invoke(cmd_work.work_plugins, ['non_existent'])
        self.assertIsNotNone(result.exception)

    def test_work_show(self):
        """Test verdi work show"""
        workchain_one = WorkCalculation().store()
        workchain_two = WorkCalculation().store()
        workchains = [workchain_one, workchain_two]

        # Running without identifiers should not except and not print anything
        options = []
        result = self.cli_runner.invoke(cmd_work.work_show, options)
        self.assertIsNone(result.exception)
        self.assertEquals(len(get_result_lines(result)), 0)

        # Giving a single identifier should print a non empty string message
        options = [str(workchain_one.pk)]
        result = self.cli_runner.invoke(cmd_work.work_show, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

        # Giving multiple identifiers should print a non empty string message
        options = [str(workchain.pk) for workchain in workchains]
        result = self.cli_runner.invoke(cmd_work.work_show, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Common utility functions for command line commands."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import os
import sys
from tabulate import tabulate


def get_env_with_venv_bin():
    """
    Create a clone of the current running environment with the AIIDA_PATH variable set to the
    value configured in the AIIDA_CONFIG_FOLDER variable
    """
    from aiida.common import setup

    currenv = os.environ.copy()
    currenv['PATH'] = os.path.dirname(sys.executable) + ':' + currenv['PATH']
    currenv['AIIDA_PATH'] = os.path.abspath(os.path.expanduser(setup.AIIDA_CONFIG_FOLDER))
    currenv['PYTHONUNBUFFERED'] = 'True'

    return currenv


def format_local_time(timestamp, format_str='%Y-%m-%d %H:%M:%S'):
    """
    Format a datetime object or UNIX timestamp in a human readable format

    :param timestamp: a datetime object or a float representing a UNIX timestamp
    :param format_str: optional string format to pass to strftime
    """
    from aiida.utils import timezone

    if isinstance(timestamp, float):
        return timezone.datetime.fromtimestamp(timestamp).strftime(format_str)

    return timestamp.strftime(format_str)


def print_last_process_state_change(process_type=None):
    """
    Print the last time that a process of the specified type has changed its state.
    This function will also print a warning if the daemon is not running.

    :param process_type: optional process type for which to get the latest state change timestamp.
        Valid process types are either 'calculation' or 'work'.
    """
    from aiida.cmdline.utils.echo import echo_info, echo_warning
    from aiida.daemon.client import get_daemon_client
    from aiida.utils import timezone
    from aiida.common.utils import str_timedelta
    from aiida.work.utils import get_process_state_change_timestamp

    client = get_daemon_client()

    timestamp = get_process_state_change_timestamp(process_type)

    if timestamp is None:
        echo_info('last time an entry changed state: never')
    else:
        timedelta = timezone.delta(timestamp, timezone.now())
        formatted = format_local_time(timestamp, format_str='at %H:%M:%S on %Y-%m-%d')
        relative = str_timedelta(timedelta, negative_to_zero=True, max_num_fields=1)
        echo_info('last time an entry changed state: {} ({})'.format(relative, formatted))

    if not client.is_daemon_running:
        echo_warning('the daemon is not running', bold=True)


def get_node_summary(node):
    """
    Return a multi line string with a pretty formatted summary of a Node

    :param node: a Node instance
    :return: a string summary of the node
    """
    from plumpy import ProcessState
    from aiida.orm.implementation.general.calculation import AbstractCalculation

    table_headers = ['Property', 'Value']
    table = []
    table.append(['type', node.__class__.__name__])
    table.append(['pk', str(node.pk)])
    table.append(['uuid', str(node.uuid)])
    table.append(['label', node.label])
    table.append(['description', node.description])
    table.append(['ctime', node.ctime])
    table.append(['mtime', node.mtime])

    if issubclass(node.__class__, AbstractCalculation):
        try:
            process_state = node.process_state
        except AttributeError:
            process_state = None

        try:
            table.append(['process state', ProcessState(process_state)])
        except ValueError:
            table.append(['process state', process_state])

        try:
            table.append(['exit status', node.exit_status])
        except AttributeError:
            table.append(['exit status', None])

    try:
        computer = node.get_computer()
    except AttributeError:
        pass
    else:
        if computer is not None:
            table.append(['computer', '[{}] {}'.format(node.get_computer().pk, node.get_computer().name)])

    try:
        code = node.get_code()
    except AttributeError:
        pass
    else:
        if code is not None:
            table.append(['code', code.label])

    return tabulate(table, headers=table_headers)


def get_node_info(node, include_summary=True):
    # pylint: disable=too-many-branches
    """
    Return a multi line string of information about the given node, such as the incoming and outcoming links

    :param include_summary: also include a summary of node properties
    :return: a string summary of the node including a description of all its links and log messages
    """
    from aiida.backends.utils import get_log_messages
    from aiida.common.links import LinkType
    from aiida.orm.calculation.work import WorkCalculation

    if include_summary:
        result = get_node_summary(node)
    else:
        result = ''

    nodes_input = node.get_inputs(link_type=LinkType.INPUT, also_labels=True)
    nodes_caller = node.get_inputs(link_type=LinkType.CALL, also_labels=True)
    nodes_create = node.get_outputs(link_type=LinkType.CREATE, also_labels=True)
    nodes_return = node.get_outputs(link_type=LinkType.RETURN, also_labels=True)
    nodes_called = node.get_outputs(link_type=LinkType.CALL, also_labels=True)
    nodes_output = nodes_create + nodes_return

    if nodes_caller:
        table = []
        table_headers = ['Called by', 'PK', 'Type']
        for key, value in nodes_caller:
            table.append([key, value.pk, value.__class__.__name__])
        result += '\n{}'.format(tabulate(table, headers=table_headers))

    if nodes_input:
        table = []
        table_headers = ['Inputs', 'PK', 'Type']
        for key, value in nodes_input:
            if key == 'code':
                continue
            table.append([key, value.pk, value.__class__.__name__])
        result += '\n{}'.format(tabulate(table, headers=table_headers))

    if nodes_output:
        table = []
        table_headers = ['Outputs', 'PK', 'Type']
        for key, value in nodes_output:
            table.append([key, value.pk, value.__class__.__name__])
        result += '\n{}'.format(tabulate(table, headers=table_headers))

    if nodes_called:
        table = []
        table_headers = ['Called', 'PK', 'Type']
        for key, value in nodes_called:
            table.append([key, value.pk, value.__class__.__name__])
        result += '\n{}'.format(tabulate(table, headers=table_headers))

    log_messages = get_log_messages(node)

    if log_messages:
        table = []
        table_headers = ['Log messages']
        table.append(['There are {} log messages for this calculation'.format(len(log_messages))])
        if isinstance(node, WorkCalculation):
            table.append(["Run 'verdi work report {}' to see them".format(node.pk)])
        else:
            table.append(["Run 'verdi calculation logshow {}' to see them".format(node.pk)])
        result += '\n{}'.format(tabulate(table, headers=table_headers))

    return result


def get_calculation_log_report(calculation):
    """
    Return a multi line string representation of the log messages and output of a given calculation

    :param calculation: the calculation node
    :return: a string representation of the log messages and scheduler output
    """
    from aiida.backends.utils import get_log_messages
    from aiida.common.datastructures import calc_states

    log_messages = get_log_messages(calculation)
    scheduler_out = calculation.get_scheduler_output()
    scheduler_err = calculation.get_scheduler_error()
    calculation_state = calculation.get_state()
    scheduler_state = calculation.get_scheduler_state()

    report = []

    if calculation_state == calc_states.WITHSCHEDULER:
        state_string = '{}, scheduler state: {}'.format(calculation_state, scheduler_state
                                                        if scheduler_state else '(unknown)')
    else:
        state_string = '{}'.format(calculation_state)

    label_string = ' [{}]'.format(calculation.label) if calculation.label else ''

    report.append('*** {}{}: {}'.format(calculation.pk, label_string, state_string))

    if scheduler_out is None:
        report.append('*** Scheduler output: N/A')
    elif scheduler_out:
        report.append('*** Scheduler output:\n{}'.format(scheduler_out))
    else:
        report.append('*** (empty scheduler output file)')

    if scheduler_err is None:
        report.append('*** Scheduler errors: N/A')
    elif scheduler_err:
        report.append('*** Scheduler errors:\n{}'.format(scheduler_err))
    else:
        report.append('*** (empty scheduler errors file)')

    if log_messages:
        report.append('*** {} LOG MESSAGES:'.format(len(log_messages)))
    else:
        report.append('*** 0 LOG MESSAGES')

    for log in log_messages:
        report.append('+-> {} at {}'.format(log['levelname'], log['time']))
        for message in log['message'].splitlines():
            report.append(' | {}'.format(message))

    return '\n'.join(report)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""The CLI for legacy workflows."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params.types import LegacyWorkflowParamType
from aiida.cmdline.params.options.multivalue import MultipleValueOption
from aiida.cmdline.params import options
from aiida.cmdline.utils.decorators import with_dbenv, deprecated_command
from aiida.cmdline.utils import echo

DEPRECATION_MSG = 'Legacy workflows have been deprecated in AiiDA v1.0.'


@verdi.group('workflow')
def verdi_workflow():
    """Inspect and manage legacy workflows."""
    pass


def format_pk(workflow):
    return '(pk: {})'.format(workflow.pk)


@verdi_workflow.command('logshow')
@click.argument('workflows', type=LegacyWorkflowParamType(), nargs=-1)
@deprecated_command(DEPRECATION_MSG)
def workflow_logshow(workflows):
    """Show the log for each workflow in a list of WORKFLOWS."""
    from aiida.backends.utils import get_log_messages
    for workflow in workflows:
        log_messages = get_log_messages(workflow)
        label_str = ' [{}]'.format(workflow.label) if workflow.label else ''
        state = workflow.get_state()
        echo.echo('*** {pk}{label}: {state}'.format(pk=format_pk(workflow), label=label_str, state=state))

        if workflow.get_report():
            echo.echo('Print the report with `verdi workflow report {}`'.format(format_pk(workflow)))
        else:
            echo.echo('*** Report is empty')

        if log_messages:
            echo.echo('*** {} LOG MESSAGES:'.format(len(log_messages)))
        else:
            echo.echo('*** NO LOG MESSAGES')

        for log in log_messages:
            echo.echo('+-> {} at {}'.format(log['levelname'], log['time']))
            echo.echo('\n'.join(['|\t{}'.format(msg) for msg in log.get('message', '').splitlines()]))


@verdi_workflow.command('kill')
@options.FORCE()
@click.option('-v', '--verbose', is_flag=True)
@click.argument('workflows', type=LegacyWorkflowParamType(), nargs=-1)
@deprecated_command(DEPRECATION_MSG)
@with_dbenv()
def workflow_kill(force, verbose, workflows):
    """Kill each in a list of WORKFLOWS (given by ID, UUID or label)"""
    from aiida.orm.workflow import WorkflowKillError, WorkflowUnkillable
    plural = '' if len(workflows) <= 1 else 's'
    if not force and workflows:
        click.confirm('Are you sure you want to kill {} workflow{}?'.format(len(workflows), plural), abort=True)
    counter = 0
    for workflow in workflows:
        try:
            workflow.kill(verbose=verbose)
            counter += 1
        except WorkflowKillError as error:
            echo.echo_error(error.__class__.__name__)
            for msg in error.error_message_list:
                echo.echo_error('\t' + msg)
        except WorkflowUnkillable as error:
            echo.echo_error(error.__class__.__name__)
            echo.echo_error(str(error))

    if verbose:
        echo.echo_success('{} workflow{} killed.'.format(counter, plural))


@verdi_workflow.command('report')
@click.argument('workflow', type=LegacyWorkflowParamType())
@deprecated_command(DEPRECATION_MSG)
def workflow_report(workflow):
    """Report on a WORKFLOW (given by ID, UUID or label)."""
    echo.echo('### WORKFLOW {} ###'.format(format_pk(workflow)))
    echo.echo('\n'.join(workflow.get_report()))


@verdi_workflow.command('list')
@click.option('-s', '--short', is_flag=True, help='short form (only subworkflows and steps, no calculations)')
@options.ALL_STATES(help='show all existing AiiDA workflows, not only running ones')
@click.option('-d', '--depth', type=int, default=16,
    help='show only steps down to a depth of DEPTH levels in subworkflows (0 means only the parent)')  # yapf: disable
@click.option(
    '-p', '--past-days', type=int, help='add a filter to show only workflows created in the past PAST_DAYS days')
@click.option('-W', '--workflows', default=[], type=LegacyWorkflowParamType(), cls=MultipleValueOption,
    help='limit the listing to these workflows')  # yapf: disable
@deprecated_command(DEPRECATION_MSG)
@with_dbenv()
def workflow_list(short, all_states, depth, past_days, workflows):
    """List legacy workflows"""
    from aiida.backends.utils import get_workflow_list
    from aiida.orm.workflow import get_workflow_info
    from aiida.orm.backend import construct_backend  # pylint: disable=no-name-in-module
    tab_size = 2

    backend = construct_backend()
    current_user = backend.users.get_automatic_user()

    wf_list = get_workflow_list(
        [workflow.pk for workflow in workflows], user=current_user, all_states=all_states, n_days_ago=past_days)
    for workflow in wf_list:
        if not workflow.is_subworkflow() or workflow in workflows:
            echo.echo('\n'.join(get_workflow_info(workflow, tab_size=tab_size, short=short, depth=depth)))

    if not workflows:
        if all_states:
            echo.echo_info('# No workflows found')
        else:
            echo.echo_info('# No running workflows found')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.testing import CliRunner
from traceback import format_exception

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands.cmd_workflow import workflow_list, workflow_kill, workflow_report, workflow_logshow, \
    format_pk


def debug_msg(result):
    return ''.join(format_exception(*result.exc_info))


def format_pk(workflow):
    return '(pk: {})'.format(workflow.pk)


def format_wf_for_list(workflow):
    return '{} {}'.format(workflow.__class__.__name__, format_pk(workflow))


class TestVerdiLegacyWorkflow(AiidaTestCase):

    @classmethod
    def setUpClass(cls):
        from aiida.workflows.test import WFTestEmpty, WFTestSimpleWithSubWF

        super(TestVerdiLegacyWorkflow, cls).setUpClass()
        cls.runner = CliRunner()

        cls.workflow = WFTestEmpty()
        cls.workflow.store()
        cls.other_workflow = WFTestEmpty()
        cls.other_workflow.store()
        cls.done_workflow = WFTestEmpty()
        cls.done_workflow.store()
        cls.done_workflow.set_state('FINISHED')
        cls.super_workflow = WFTestSimpleWithSubWF()
        cls.super_workflow.store()
        cls.super_workflow.start()

    def test_workflow_list_default(self):
        result = self.runner.invoke(workflow_list, [])
        self.assertIsNone(result.exception, msg=debug_msg(result))
        self.assertIn(format_wf_for_list(self.workflow), result.output)
        self.assertIn(format_wf_for_list(self.other_workflow), result.output)
        self.assertNotIn(format_wf_for_list(self.done_workflow), result.output)

    def test_workflow_list_workflows(self):
        result = self.runner.invoke(workflow_list, ['--workflows={}'.format(self.workflow.pk)])
        self.assertIsNone(result.exception, msg=debug_msg(result))
        self.assertIn(format_wf_for_list(self.workflow), result.output)
        self.assertNotIn(format_wf_for_list(self.other_workflow), result.output)
        self.assertNotIn(format_wf_for_list(self.done_workflow), result.output)

    def test_workflow_list_states(self):
        result = self.runner.invoke(workflow_list, ['--all-states'])
        self.assertIsNone(result.exception, msg=debug_msg(result))
        self.assertIn(format_wf_for_list(self.workflow), result.output)
        self.assertIn(format_wf_for_list(self.other_workflow), result.output)
        self.assertIn(format_wf_for_list(self.done_workflow), result.output)

    def test_workflow_list_depth(self):
        results = []
        results.append(self.runner.invoke(workflow_list, ['--depth=0']))
        results.append(self.runner.invoke(workflow_list, ['--depth=1']))
        results.append(self.runner.invoke(workflow_list, ['--depth=2']))

        for result in results:
            self.assertIsNone(result.exception, msg=debug_msg(result))

        self.assertTrue(len(results[0].output) < len(results[1].output))
        self.assertTrue(len(results[1].output) < len(results[2].output))

    def test_workflow_report(self):
        result = self.runner.invoke(workflow_report, [str(self.super_workflow.uuid)])
        self.assertIsNone(result.exception, msg=debug_msg(result))
        self.assertIn(format_pk(self.super_workflow), result.output)

    def test_workflow_kill(self):
        from aiida.workflows.test import WFTestEmpty
        running_workflow = WFTestEmpty()
        running_workflow.store()
        result = self.runner.invoke(workflow_kill, [str(running_workflow.uuid)], input='y\n')
        self.assertIsNone(result.exception)
        self.assertNotIn('WorkflowKillError', result.output)
        self.assertNotIn('WorkflowUnkillable', result.output)
        self.assertEqual(running_workflow.get_state(), 'FINISHED')

        self.assertEqual(self.done_workflow.get_state(), 'FINISHED')
        result = self.runner.invoke(workflow_kill, [str(self.done_workflow.pk)], input='y\n')
        self.assertIsNone(result.exception)
        self.assertIn('WorkflowUnkillable', result.output)
        self.assertEqual(self.done_workflow.get_state(), 'FINISHED')

    def test_workflow_logshow(self):
        result = self.runner.invoke(workflow_logshow, [str(self.super_workflow.pk)])
        self.assertIsNone(result.exception, msg=debug_msg(result))
        self.assertIn(format_pk(self.super_workflow), result.output)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Generic functions to verify the integrity of the database and optionally apply patches to fix problems."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import uuid as UUID

from aiida.common.exceptions import IntegrityError


def verify_node_uuid_uniqueness():
    """Check whether the database contains nodes with duplicate UUIDS.

    :raises: IntegrityError if database contains nodes with duplicate UUIDS.
    """
    duplicates = get_duplicate_node_uuids()

    if duplicates:
        raise IntegrityError('your database contains nodes with duplicate UUIDS: '
                             'run `verdi database integrity duplicate-node-uuid` to return to a consistent state')


def get_duplicate_node_uuids():
    """Retrieve nodes with duplicate UUIDS.

    :return: list of tuples of (pk, uuid) of nodes with duplicate UUIDs
    """
    from aiida.orm.backend import construct_backend

    backend = construct_backend()
    duplicates = backend.query_manager.get_duplicate_node_uuids()

    return duplicates


def deduplicate_node_uuids(dry_run=True):
    """Detect and solve nodes with duplicate UUIDs.

    Before aiida-core v1.0.0, there was no uniqueness constraint on the UUID column of the Node table in the database.
    This made it possible to store multiple nodes with identical UUIDs without the database complaining. This was bug
    was fixed in aiida-core=1.0.0 by putting an explicit uniqueness constraint on node UUIDs on the database level.
    However, this would leave databases created before this patch with duplicate UUIDs in an inconsistent state. This
    command will run an analysis to detect duplicate UUIDs in the node table and solve it by generating new UUIDs. Note
    that it will not delete or merge any nodes.

    :param dry_run: when True, no actual changes will be made
    :return: list of strings denoting the performed operations, or those that would have been applied for dry_run=False
    """
    from collections import defaultdict

    from aiida.backends.settings import AIIDANODES_UUID_VERSION
    from aiida.orm import load_node

    uuid_generator = getattr(UUID, 'uuid{}'.format(AIIDANODES_UUID_VERSION))

    mapping = defaultdict(list)

    for pk, uuid in get_duplicate_node_uuids():
        mapping[uuid].append(int(pk))

    def copy_repo_folder(node_source, uuid):
        """
        Copy the repository folder from source node to a new location based on the given UUID.

        :param node_source: the node whose repository folder contents to copy
        :param uuid: the UUID that will be used to generate the sharded folder location for the copied folder
        """
        from aiida.common.folders import RepositoryFolder
        folder = RepositoryFolder('node', uuid)
        folder.replace_with_folder(node_source.folder.abspath)

    messages = []

    for uuid, nodes in mapping.items():

        uuid_old = None

        for pk in nodes:

            # We don't have to change all nodes that have the same UUID, the first one can keep the original
            if uuid_old is None:
                node_source = load_node(pk)
                uuid_old = uuid
                continue

            node = load_node(pk)
            uuid_new = str(uuid_generator())

            if dry_run:
                messages.append('would update UUID of Node<{}> from {} to {}'.format(pk, uuid_old, uuid_new))
            else:
                node._dbnode.uuid = uuid_new  # pylint: disable=protected-access
                node._dbnode.save()  # pylint: disable=protected-access
                copy_repo_folder(node_source, uuid_new)
                messages.append('updated UUID of Node<{}> from {} to {}'.format(pk, uuid_old, uuid_new))

    if not messages:
        messages = ['no duplicate UUIDs found']

    return messages



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from copy import copy
import six
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.session import make_transient
from sqlalchemy.orm.attributes import flag_modified

from aiida.common.utils import type_check
from aiida.orm.computer import Computer, ComputerCollection
from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.common.exceptions import (ConfigurationError, InvalidOperation)
from aiida.common.lang import override
from . import utils
import aiida.utils.json as json


class SqlaComputerCollection(ComputerCollection):
    def create(self, name, hostname, description='', transport_type='', scheduler_type='', workdir=None,
               enabled_state=True):
        return SqlaComputer(self.backend, name=name, hostname=hostname, description=description,
                            transport_type=transport_type, scheduler_type=scheduler_type, workdir=workdir,
                            enabled=enabled_state)

    def list_names(cls):
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()
        return session.query(DbComputer.name).all()

    def delete(self, id):
        import aiida.backends.sqlalchemy
        try:
            session = aiida.backends.sqlalchemy.get_scoped_session()
            session.query(DbComputer).get(id).delete()
            session.commit()
        except SQLAlchemyError as exc:
            raise InvalidOperation("Unable to delete the requested computer: it is possible that there "
                                   "is at least one node using this computer (original message: {})".format(exc))

    def from_dbmodel(self, computer):
        """
        Construct a SqlaComputer instance from the corresponding database entry

        :param computer: The DbComputer instance
        :return: The Computer instance
        :rtype: :class:`aiida.orm.implementation.sqlalchemy.computer.SqlaComputer`
        """
        return SqlaComputer.from_dbmodel(computer, self.backend)


class SqlaComputer(Computer):
    @classmethod
    def from_dbmodel(cls, dbmodel, backend):
        type_check(dbmodel, DbComputer)
        computer = SqlaComputer.__new__(cls)
        super(SqlaComputer, computer).__init__(backend)
        computer._dbcomputer = utils.ModelWrapper(dbmodel)
        return computer

    @property
    def uuid(self):
        return six.text_type(self._dbcomputer.uuid)

    @property
    def pk(self):
        return self._dbcomputer.id

    @property
    def id(self):
        return self._dbcomputer.id

    def __init__(self, backend, name, hostname, description='', transport_type='', scheduler_type='', workdir=None,
                 enabled=True):
        super(Computer, self).__init__(backend)
        self._dbcomputer = utils.ModelWrapper(DbComputer(
            name=name,
            hostname=hostname,
            description=description,
            transport_type=transport_type,
            scheduler_type=scheduler_type,
            enabled=enabled
        ))
        if workdir:
            self.set_workdir(workdir)

    def set(self, **kwargs):

        is_modified = False

        for key, val in kwargs.items():
            if hasattr(self._dbcomputer, key):
                setattr(self._dbcomputer, key, val)
            else:
                self._dbcomputer._metadata[key] = val
                is_modified = True

        if is_modified:
            flag_modified(self._dbcomputer, "_metadata")

    @property
    def full_text_info(self):
        ret_lines = []
        ret_lines.append("Computer name:     {}".format(self.name))
        ret_lines.append(" * PK:             {}".format(self.pk))
        ret_lines.append(" * UUID:           {}".format(self.uuid))
        ret_lines.append(" * Description:    {}".format(self.description))
        ret_lines.append(" * Hostname:       {}".format(self.hostname))
        ret_lines.append(" * Enabled:        {}".format("True" if self.is_enabled() else "False"))
        ret_lines.append(" * Transport type: {}".format(self.get_transport_type()))
        ret_lines.append(" * Scheduler type: {}".format(self.get_scheduler_type()))
        ret_lines.append(" * Work directory: {}".format(self.get_workdir()))
        ret_lines.append(" * Shebang:        {}".format(self.get_shebang()))
        ret_lines.append(" * mpirun command: {}".format(" ".join(
            self.get_mpirun_command())))
        def_cpus_machine = self.get_default_mpiprocs_per_machine()

        if def_cpus_machine is not None:
            ret_lines.append(" * Default number of cpus per machine: {}".format(
                def_cpus_machine))
        ret_lines.append(" * Used by:        {} nodes".format(
            len(self._dbcomputer.dbnodes)))

        ret_lines.append(" * prepend text:")

        if self.get_prepend_text().strip():
            for l in self.get_prepend_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No prepend text.")

        ret_lines.append(" * append text:")

        if self.get_append_text().strip():
            for l in self.get_append_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No append text.")

        return "\n".join(ret_lines)

    @property
    def is_stored(self):
        return self._dbcomputer.id is not None

    @override
    def copy(self):
        from aiida.backends.sqlalchemy import get_scoped_session
        session = get_scoped_session()

        if not self.is_stored:
            raise InvalidOperation("You can copy a computer only after having stored it")

        newdbcomputer = copy(self._dbcomputer)
        make_transient(newdbcomputer)
        session.add(newdbcomputer)

        newobject = self.__class__(dbcomputer=newdbcomputer)

        return newobject

    @property
    def dbcomputer(self):
        return self._dbcomputer._model

    def store(self):
        self.validate()

        try:
            self._dbcomputer.save()
        except SQLAlchemyError:
            raise ValueError(
                "Integrity error, probably the hostname already exists in the"
                " DB")

        return self

    @property
    def name(self):
        return self._dbcomputer.name

    @property
    def description(self):
        return self._dbcomputer.description

    @property
    def hostname(self):
        return self._dbcomputer.hostname

    def _get_metadata(self):
        return self._dbcomputer._metadata

    def _set_metadata(self, metadata_dict):
        self._dbcomputer._metadata = metadata_dict
        flag_modified(self._dbcomputer, "_metadata")

    def get_transport_params(self):
        """
        Return transport params stored in dbcomputer instance
        """
        return self._dbcomputer.transport_params

    def set_transport_params(self, val):
        try:
            json.dumps(val)  # Check if json compatible
            self._dbcomputer.transport_params = val
        except ValueError:
            raise ValueError("The set of transport_params are not JSON-able")

    def get_workdir(self):
        try:
            return self._dbcomputer.get_workdir()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "/scratch/{username}/aiida_run/"

    def set_workdir(self, val):
        pass
        metadata = self._get_metadata()
        metadata['workdir'] = val
        self._set_metadata(metadata)

    def get_shebang(self):
        try:
            return self._dbcomputer.get_shebang()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "#!/bin/bash"

    def get_name(self):
        return self._dbcomputer.name

    def set_name(self, val):
        self._dbcomputer.name = val

    def get_hostname(self):
        return self._dbcomputer.hostname

    def set_hostname(self, val):
        self._dbcomputer.hostname = val

    def get_description(self):
        return self._dbcomputer.description

    def set_description(self, val):
        self._dbcomputer.description = val

    def get_calculations_on_computer(self):
        from aiida.backends.sqlalchemy.models.node import DbNode
        return DbNode.query.filter(
            DbNode.dbcomputer_id == self._dbcomputer.id,
            DbNode.type.like("calculation%")).all()

    def is_enabled(self):
        return self._dbcomputer.enabled

    def set_enabled_state(self, enabled):
        self._dbcomputer.enabled = enabled

    def get_scheduler_type(self):
        return self._dbcomputer.scheduler_type

    def set_scheduler_type(self, val):

        self._dbcomputer.scheduler_type = val

    def get_transport_type(self):
        return self._dbcomputer.transport_type

    def set_transport_type(self, val):
        self._dbcomputer.transport_type = val



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi devel`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import tempfile

from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_devel
from aiida.common import setup as aiida_cfg
from aiida.common.setup import _property_table, exists_property, get_property, del_property


def get_result_lines(result):
    return [e for e in result.output.split('\n') if e]


class TestVerdiDevel(AiidaTestCase):
    """Tests for `verdi devel`."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiDevel, cls).setUpClass()

        cls._old_aiida_config_folder = None
        cls._new_aiida_config_folder = tempfile.mkdtemp()
        cls._old_aiida_config_folder = aiida_cfg.AIIDA_CONFIG_FOLDER

        aiida_cfg.AIIDA_CONFIG_FOLDER = cls._new_aiida_config_folder
        aiida_cfg.create_base_dirs()

        cls.profile_name = 'aiida_dummy_profile_1234'
        cls.dummy_profile = {}
        cls.dummy_profile['backend'] = 'django'
        cls.dummy_profile['db_host'] = 'localhost'
        cls.dummy_profile['db_port'] = '5432'
        cls.dummy_profile['email'] = 'dummy@localhost'
        cls.dummy_profile['db_name'] = cls.profile_name
        cls.dummy_profile['db_user'] = 'dummy_user'
        cls.dummy_profile['db_pass'] = 'dummy_pass'
        cls.dummy_profile['repo'] = aiida_cfg.AIIDA_CONFIG_FOLDER + '/repository_' + cls.profile_name

        aiida_cfg.create_config_noninteractive(profile=cls.profile_name, **cls.dummy_profile)
        aiida_cfg.set_default_profile(cls.profile_name, force_rewrite=True)

    @classmethod
    def tearDownClass(cls, *args, **kwargs):
        """
        Remove dummy profiles created in setUpClass.

        :param args: list of arguments
        :param kwargs: list of keyword arguments
        """
        import os
        import shutil

        aiida_cfg.AIIDA_CONFIG_FOLDER = cls._old_aiida_config_folder

        if os.path.isdir(cls._new_aiida_config_folder):
            shutil.rmtree(cls._new_aiida_config_folder)

    def setUp(self):
        self.runner = CliRunner()

    def tearDown(self):
        """Delete any properties that were set."""
        for prop in sorted(_property_table.keys()):
            if exists_property(prop):
                del_property(prop)

    def test_describe_properties(self):
        """Test the `verdi devel describeproperties` command"""
        result = self.runner.invoke(cmd_devel.devel_describeproperties, [])

        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) > 0)

        for prop in _property_table:
            self.assertIn(prop, result.output)

    def test_list_properties(self):
        """Test the `verdi devel listproperties` command"""

        # Since dummy configuration does not have explicit properties set, with default options the result is empty
        options = []
        result = self.runner.invoke(cmd_devel.devel_listproperties, options)
        self.assertIsNone(result.exception)
        self.assertTrue(len(get_result_lines(result)) == 0)

        # Toggling the all flag should show all available options
        for flag in ['-a', '--all']:

            options = [flag]
            result = self.runner.invoke(cmd_devel.devel_listproperties, options)
            self.assertIsNone(result.exception)
            self.assertTrue(len(get_result_lines(result)) > 0)

            for prop in _property_table:
                self.assertIn(prop, result.output)

    def test_set_property(self):
        """Test the `verdi devel setproperty` command."""
        property_name = 'daemon.timeout'
        property_values = [10, 20]

        for property_value in property_values:
            options = [property_name, str(property_value)]
            result = self.runner.invoke(cmd_devel.devel_setproperty, options)

            self.assertIsNone(result.exception)
            self.assertEquals(get_property(property_name), property_value)

    def test_get_property(self):
        """Test the `verdi devel getproperty` command."""
        property_name = 'daemon.timeout'
        property_value = 30

        options = [property_name, str(property_value)]
        result = self.runner.invoke(cmd_devel.devel_setproperty, options)

        options = [property_name]
        result = self.runner.invoke(cmd_devel.devel_getproperty, options)
        self.assertIsNone(result.exception)
        self.assertEquals(str(property_value), result.output.strip())

    def test_del_property(self):
        """Test the `verdi devel delproperty` command."""
        property_name = 'daemon.timeout'
        property_value = 30

        options = [property_name, str(property_value)]
        result = self.runner.invoke(cmd_devel.devel_setproperty, options)

        options = [property_name]
        result = self.runner.invoke(cmd_devel.devel_delproperty, options)
        self.assertIsNone(result.exception)
        self.assertIn(property_name, result.output.strip())
        self.assertFalse(exists_property(property_name))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions for command line commands operating on the repository."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import io

from aiida.cmdline.utils import echo


def cat_repo_files(node, path):
    """
    Given a Node and a relative path to a file in the Node repository directory,
    prints in output the content of the file.

    :param node: a Node instance
    :param path: a string with the relative path to list. Must be a file.
    :raise ValueError: if the file is not found, or is a directory.
    """
    import os
    import sys

    fldr = node.folder

    is_dir = False
    parts = path.split(os.path.sep)
    # except the last item
    for item in parts[:-1]:
        fldr = fldr.get_subfolder(item)
    if parts:
        if fldr.isdir(parts[-1]):
            fldr = fldr.get_subfolder(parts[-1])
            is_dir = True
        else:
            fname = parts[-1]
    else:
        is_dir = True

    if is_dir:
        if not fldr.isdir('.'):
            raise ValueError("No directory '{}' in the repo".format(path))
        else:
            raise ValueError("'{}' is a directory".format(path))
    else:
        if not fldr.isfile(fname):
            raise ValueError("No file '{}' in the repo".format(path))

        absfname = fldr.get_abs_path(fname)
        with io.open(absfname, encoding='utf8') as repofile:
            for line in repofile:
                sys.stdout.write(line)


def list_repo_files(node, path, color):
    """
    Given a Node and a relative path prints in output the list of files
    in the given path in the Node repository directory.

    :param node: a Node instance
    :param path: a string with the relative path to list. Can be a file.
    :param color: boolean, if True prints with the codes to show colors.
    :raise ValueError: if the file or directory is not found.
    """
    import os

    fldr = node.folder

    is_dir = False
    parts = path.split(os.path.sep)
    # except the last item
    for item in parts[:-1]:
        fldr = fldr.get_subfolder(item)
    if parts:
        if fldr.isdir(parts[-1]):
            fldr = fldr.get_subfolder(parts[-1])
            is_dir = True
        else:
            fname = parts[-1]
    else:
        is_dir = True

    if is_dir:
        if not fldr.isdir('.'):
            raise ValueError("{}: No such file or directory in the repo".format(path))

        for elem, elem_is_file in sorted(fldr.get_content_list(only_paths=False)):
            if elem_is_file or not color:
                echo.echo(elem)
            else:
                # BOLD("1;") and color 34=blue
                outstr = "\x1b[1;{color_id}m{elem}\x1b[0m".format(color_id=34, elem=elem)
                echo.echo(outstr)
    else:
        if not fldr.isfile(fname):
            raise ValueError("{}: No such file or directory in the repo".format(path))

        echo.echo(fname)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
class ErrorAccumulator(object):
    """
    Allows to run a number of functions and collect all the errors they raise

    This allows to validate multiple things and tell the user about all the
    errors encountered at once. Works best if the individual functions do not depend on each other.

    Does not allow to trace the stack of each error, therefore do not use for debugging, but for
    semantical checking with user friendly error messages.
    """

    def __init__(self, *error_cls):
        self.error_cls = error_cls
        self.errors = {k: [] for k in self.error_cls}

    def run(self, function, *args, **kwargs):
        try:
            function(*args, **kwargs)
        except self.error_cls as err:
            self.errors[err.__class__].append(err)

    def success(self):
        return bool(not any(self.errors.values()))

    def result(self, raise_error=Exception):
        if raise_error:
            self.raise_errors(raise_error)
        return self.success(), self.errors

    def raise_errors(self, raise_cls):
        if not self.success():
            raise raise_cls('The following errors were encountered: {}'.format(self.errors))




# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import 
import click
from . import validators


class overridable_option(object):
    """
    Wrapper around click option that allows to store the name
    and some defaults but also to override them later, for example
    to change the help message for a certain command.
    """

    def __init__(self, *args, **kwargs):
        """
        Store the defaults
        """
        self.args = args
        self.kwargs = kwargs

    def __call__(self, *args, **kwargs):
        """
        Override args if passed and update the kwargs and return the click.option
        If a callback is specified in the kwargs make sure to bind the callback_kwargs
        to it using the partial lambda construct of functools
        """
        import functools

        if not args:
            args_copy = self.args
        else:
            args_copy = args

        kw_copy = self.kwargs.copy()
        kw_copy.update(kwargs)

        # Pop the optional callback_kwargs if present
        callback_kwargs = kw_copy.pop('callback_kwargs', {})

        if 'callback' in kw_copy:
            callback_plain = kw_copy['callback']
            callback_bound = functools.partial(callback_plain, callback_kwargs)
            kw_copy['callback'] = callback_bound

        return click.option(*args_copy, **kw_copy)


code = overridable_option(
    '-c', '--code', type=click.STRING, required=True,
    callback=validators.validate_code,
    help='the label of the AiiDA code object to use'
)


structure = overridable_option(
    '-s', '--structure', type=click.INT, required=True,
    callback=validators.validate_structure,
    help='the pk of a structure node'
)


pseudo_family = overridable_option(
    '-p', '--pseudo-family', type=click.STRING, required=True,
    callback=validators.validate_pseudo_family,
    help='the name of the pseudo potential family to use'
)


kpoint_mesh = overridable_option(
    '-k', '--kpoint-mesh', 'kpoints', nargs=3, type=click.INT, default=[2, 2, 2], show_default=True,
    callback=validators.validate_kpoint_mesh,
    help='the number of points in the kpoint mesh along each basis vector'
)


calculation = overridable_option(
    '-C', '--calculation', type=click.INT, required=True,
    callback=validators.validate_calculation,
    help='the pk of a calculation node'
)


max_num_machines = overridable_option(
    '-m', '--max-num-machines', type=click.INT, default=1, show_default=True,
    help='the maximum number of machines (nodes) to use for the calculations'
)


max_wallclock_seconds = overridable_option(
    '-w', '--max-wallclock-seconds', type=click.INT, default=1800, show_default=True,
    help='the maximum wallclock time in seconds to set for the calculations'
)


daemon = overridable_option(
    '-d', '--daemon', is_flag=True, default=False, show_default=True,
    help='submit the workchain to the daemon instead of running it locally'
)


group = overridable_option(
    '-g', '--group', type=click.STRING, required=True,
    callback=validators.validate_group,
    help='the name or pk of a Group'
)


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Legacy workflow parameter type."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from .identifier import IdentifierParamType


class LegacyWorkflowParamType(IdentifierParamType):
    """The ParamType for identifying legacy workflows."""

    name = 'LegacyWorkflow'

    @property
    def orm_class_loader(self):
        """Return the orm entity loader class which should be a subclass of OrmEntityLoader."""
        from aiida.orm.utils.loaders import LegacyWorkflowLoader
        return LegacyWorkflowLoader



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.orm.implementation.calculation import Calculation
from aiida.orm.mixins import FunctionCalculationMixin


class InlineCalculation(FunctionCalculationMixin, Calculation):
    """
    Calculation node to record the results of a function that was run wrapped by the
    :class:`aiida.orm.calculation.inline.make_inline` decorator
    """

    _cacheable = True

    def get_desc(self):
        """
        Returns a string with the name of function that was wrapped by the make_inline decorator
        that resulted in the creation of this InlineCalculation instance

        :return: description string or None if no function name can be retrieved from the node
        """
        if self.function_name is not None:
            return '{}()'.format(self.function_name)
        else:
            return None



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Tests for `verdi process`."""
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.cmdline.commands import cmd_process
from aiida.work import runners, rmq, test_utils


def get_result_lines(result):
    return [e for e in result.output.split('\n') if e]


class TestVerdiProcess(AiidaTestCase):
    """Tests for `verdi process`."""

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        super(TestVerdiProcess, cls).setUpClass(*args, **kwargs)
        rmq_config = rmq.get_rmq_config()

        # These two need to share a common event loop otherwise the first will never send
        # the message while the daemon is running listening to intercept
        cls.runner = runners.Runner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)

        cls.daemon_runner = runners.DaemonRunner(rmq_config=rmq_config, rmq_submit=True, poll_interval=0.)

    def setUp(self):
        super(TestVerdiProcess, self).setUp()
        self.cli_runner = CliRunner()

    def test_pause_play_kill(self):
        """
        Test the pause/play/kill commands
        """
        from concurrent.futures import ThreadPoolExecutor

        calc = self.runner.submit(test_utils.WaitProcess)
        executor = ThreadPoolExecutor(max_workers=1)

        try:
            _ = executor.submit(self.daemon_runner.start)

            self.assertFalse(calc.paused)
            result = self.cli_runner.invoke(cmd_process.process_pause, [str(calc.pk)])

            self.assertTrue(calc.paused)
            self.assertIsNone(result.exception)

            result = self.cli_runner.invoke(cmd_process.process_play, [str(calc.pk)])

            self.assertFalse(calc.paused)
            self.assertIsNone(result.exception)

            result = self.cli_runner.invoke(cmd_process.process_kill, [str(calc.pk)])

            self.assertTrue(calc.is_terminated)
            self.assertTrue(calc.is_killed)
            self.assertIsNone(result.exception)
        finally:
            self.daemon_runner.stop()
            executor.shutdown()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from aiida.backends.testbase import AiidaTestCase
from aiida.common.datastructures import _sorted_datastates, calc_states, is_progressive_state_change


class TestCommonDataStructures(AiidaTestCase):

    def test_is_progressive_state_change(self):
        """Test the `is_progressive_state_change` utility function by testing all possible state change permutations."""
        for i, state_one in enumerate(_sorted_datastates):

            for j, state_two in enumerate(_sorted_datastates):

                # States will be equal and should not count as progressive state change
                if i == j:
                    self.assertFalse(is_progressive_state_change(state_one, state_two))
                elif i > j:
                    self.assertFalse(is_progressive_state_change(state_one, state_two))
                elif i < j:
                    self.assertTrue(is_progressive_state_change(state_one, state_two))
                else:
                    assert True, 'we broke math'

    def test_is_progressive_state_change_invalid_states(self):
        """Test `is_progressive_state_change` function should raise ValueError for invalid states."""
        with self.assertRaises(ValueError):
            is_progressive_state_change('NOTEXISTENT', calc_states.NEW)

        with self.assertRaises(ValueError):
            is_progressive_state_change(calc_states.NEW, 'NOTEXISTENT')



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
import collections
import six
from django.db import IntegrityError, transaction

from aiida.backends.djsite.db.models import DbComputer
from aiida.common.utils import type_check
from aiida.common.exceptions import (NotExistent, ConfigurationError,
                                     InvalidOperation, DbContentError)
from aiida.orm.computer import Computer, ComputerCollection
from . import utils
import aiida.utils.json as json



class DjangoComputerCollection(ComputerCollection):
    def create(self, name, hostname, description='', transport_type='', scheduler_type='', workdir=None,
               enabled_state=True):
        return DjangoComputer(self.backend, name=name, hostname=hostname, description=description,
                              transport_type=transport_type, scheduler_type=scheduler_type, workdir=workdir,
                              enabled=enabled_state)

    def list_names(self):
        from aiida.backends.djsite.db.models import DbComputer
        return list(DbComputer.objects.filter().values_list('name', flat=True))

    def delete(self, id):
        """ Delete the computer with the given id """
        from django.db.models.deletion import ProtectedError
        try:
            DbComputer.objects.filter(pk=id).delete()
        except ProtectedError:
            raise InvalidOperation("Unable to delete the requested computer: there"
                                   "is at least one node using this computer")

    def from_dbmodel(self, computer):
        """
        Create a DjangoComputer from a DbComputer instance

        :param computer: The dbcomputer instance
        :type computer: :class:`aiida.backends.djsite.db.models.DbComputer`
        :return: The DjangoComputer instance
        :rtype: :class:`aiida.orm.implementation.django.computer.DjangoComputer`
        """
        return DjangoComputer.from_dbmodel(computer, self.backend)


class DjangoComputer(Computer):
    @classmethod
    def from_dbmodel(cls, dbmodel, backend):
        """
        Create a DjangoUser from a dbmodel instance

        :param dbmodel: The dbmodel instance
        :type dbmodel: :class:`aiida.backends.djsite.db.models.DbComputer`
        :param backend: The backend
        :type backend: :class:`aiida.orm.implementation.django.backend.DjangoBackend`
        :return: A DjangoComputer instance
        :rtype: :class:`aiida.orm.implementation.django.computer.DjangoComputer`
        """
        type_check(dbmodel, DbComputer)
        computer = cls.__new__(cls)
        super(DjangoComputer, computer).__init__(backend)
        computer._dbcomputer = utils.ModelWrapper(dbmodel)
        return computer

    @property
    def uuid(self):
        return six.text_type(self._dbcomputer.uuid)

    @property
    def pk(self):
        return self._dbcomputer.pk

    @property
    def id(self):
        return self._dbcomputer.pk

    def __init__(self, backend, name, hostname, description='', transport_type='', scheduler_type='', workdir=None,
                 enabled=True):
        super(Computer, self).__init__(backend)
        self._dbcomputer = utils.ModelWrapper(DbComputer(
            name=name,
            hostname=hostname,
            description=description,
            transport_type=transport_type,
            scheduler_type=scheduler_type,
            enabled=enabled
        ))
        if workdir:
            self.set_workdir(workdir)

    def set(self, **kwargs):
        for k, v in kwargs.items():
            try:
                method = getattr(self, 'set_{}'.format(k))
            except AttributeError:
                raise ValueError("Unable to set '{0}', no set_{0} method "
                                 "found".format(k))
            if not isinstance(method, collections.Callable):
                raise ValueError("Unable to set '{0}', set_{0} is not "
                                 "callable!".format(k))
            method(v)

    @property
    def full_text_info(self):
        ret_lines = []
        ret_lines.append("Computer name:     {}".format(self.name))
        ret_lines.append(" * PK:             {}".format(self.pk))
        ret_lines.append(" * UUID:           {}".format(self.uuid))
        ret_lines.append(" * Description:    {}".format(self.description))
        ret_lines.append(" * Hostname:       {}".format(self.hostname))
        ret_lines.append(" * Enabled:        {}".format(
            "True" if self.is_enabled() else "False"))
        ret_lines.append(
            " * Transport type: {}".format(self.get_transport_type()))
        ret_lines.append(
            " * Scheduler type: {}".format(self.get_scheduler_type()))
        ret_lines.append(" * Work directory: {}".format(self.get_workdir()))
        ret_lines.append(" * Shebang:        {}".format(self.get_shebang()))
        ret_lines.append(" * mpirun command: {}".format(" ".join(
            self.get_mpirun_command())))
        def_cpus_machine = self.get_default_mpiprocs_per_machine()
        if def_cpus_machine is not None:
            ret_lines.append(" * Default number of cpus per machine: {}".format(
                def_cpus_machine))
        ret_lines.append(" * Used by:        {} nodes".format(
            len(self._dbcomputer.dbnodes.all())))

        ret_lines.append(" * prepend text:")
        if self.get_prepend_text().strip():
            for l in self.get_prepend_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No prepend text.")
        ret_lines.append(" * append text:")
        if self.get_append_text().strip():
            for l in self.get_append_text().split('\n'):
                ret_lines.append("   {}".format(l))
        else:
            ret_lines.append("   # No append text.")

        return "\n".join(ret_lines)

    @property
    def is_stored(self):
        return self._dbcomputer.pk is not None

    def copy(self):
        from aiida.backends.djsite.db.models import DbComputer
        if not self.is_stored:
            raise InvalidOperation(
                "You can copy a computer only after having stored it")
        newdbcomputer = DbComputer.objects.get(pk=self.pk)
        newdbcomputer.pk = None

        newobject = self.__class__(dbcomputer=newdbcomputer)

        return newobject

    @property
    def dbcomputer(self):
        """
        Return the underlying DbComputer

        :return: Return the DbComputer
        :rtype: :class:`aiida.backends.djsite.db.models.DbComputer`
        """
        return self._dbcomputer._model

    def store(self):
        # As a first thing, I check if the data is valid
        self.validate()
        sid = transaction.savepoint()
        try:
            # transactions are needed here for Postgresql:
            # https://docs.djangoproject.com/en/1.5/topics/db/transactions/#handling-exceptions-within-postgresql-transactions
            self._dbcomputer.save()
            transaction.savepoint_commit(sid)
        except IntegrityError:
            transaction.savepoint_rollback(sid)
            raise ValueError(
                "Integrity error, probably the hostname already exists in the"
                " DB")

        return self

    @property
    def name(self):
        return self._dbcomputer.name

    @property
    def description(self):
        return self._dbcomputer.description

    @property
    def hostname(self):
        return self._dbcomputer.hostname

    def _get_metadata(self):
        return json.loads(self._dbcomputer.metadata)

    def _set_metadata(self, metadata_dict):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.metadata = json.dumps(metadata_dict)

    def get_transport_params(self):
        try:
            return json.loads(self._dbcomputer.transport_params)
        except ValueError:
            raise DbContentError(
                "Error while reading transport_params for computer {}".format(
                    self.hostname))

    def set_transport_params(self, val):
        # When setting, use the uncached _dbcomputer
        try:
            self._dbcomputer.transport_params = json.dumps(val)
        except ValueError:
            raise ValueError("The set of transport_params are not JSON-able")

    def get_workdir(self):
        try:
            return self._dbcomputer.get_workdir()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "/scratch/{username}/aiida_run/"

    def get_shebang(self):
        try:
            return self._dbcomputer.get_shebang()
        except ConfigurationError:
            # This happens the first time: I provide a reasonable default value
            return "#!/bin/bash"

    def set_workdir(self, val):
        if not isinstance(val, six.string_types):
            raise ValueError("Computer work directory needs to be string, got {}".format(val))

        metadata = self._get_metadata()
        metadata['workdir'] = val
        self._set_metadata(metadata)

    def get_name(self):
        return self._dbcomputer.name

    def set_name(self, val):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.name = val

    def get_hostname(self):
        return self._dbcomputer.hostname

    def set_hostname(self, val):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.hostname = val

    def get_description(self):
        return self._dbcomputer.description

    def set_description(self, val):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.description = val

    def get_calculations_on_computer(self):
        from aiida.backends.djsite.db.models import DbNode
        return DbNode.objects.filter(dbcomputer__name=self.name,
                                     type__startswith='calculation')

    def is_enabled(self):
        return self._dbcomputer.enabled

    def set_enabled_state(self, enabled):
        """
        Set the enabled state.

        :param enabled: the new state
        """
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.enabled = enabled

    def get_scheduler_type(self):
        return self._dbcomputer.scheduler_type

    def set_scheduler_type(self, val):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.scheduler_type = val

    def get_transport_type(self):
        return self._dbcomputer.transport_type

    def set_transport_type(self, val):
        # When setting, use the uncached _dbcomputer
        self._dbcomputer.transport_type = val
